import pandas as pd
import numpy as np
from typing import Dict, List, Optional, Union, Any

from ..metrics.average_precision import AveragePrecision
from gesund.core._utils import ValidationUtils


class PlotAveragePrecision:
    def __init__(
        self,
        class_mappings,
        coco_,
        meta_data_dict=None,
    ):
        self.class_mappings = class_mappings
        self.coco_ = coco_

        if bool(meta_data_dict):
            self.is_meta_exists = True
            meta_df = pd.DataFrame(meta_data_dict).T
            self.validation_utils = ValidationUtils(meta_df)

    def _plot_performance_by_iou_threshold(
        self, threshold: float, return_points: bool = False
    ) -> Dict[str, Any]:
        """
        Plot performance metrics at specific IoU threshold.

        :param threshold: IoU threshold value
        :type threshold: float
        :param return_points: Whether to return coordinate points
        :type return_points: bool
        :return: Dictionary containing performance metrics data
        :rtype: Dict[str, Any]
        """
        payload_dict = dict()
        payload_dict["type"] = "mixed"

        average_precision = AveragePrecision(
            class_mappings=self.class_mappings,
            coco_=self.coco_,
        )

        metrics = average_precision.calculate_ap_metrics(threshold=threshold)
        coordinates = average_precision.calculate_iou_threshold_graph(
            threshold=threshold
        )

        response = average_precision.calculate_highlighted_overall_metrics(threshold)

        if return_points:
            payload_dict["data"] = {
                "metrics": metrics,
                "coordinates": coordinates,
            }
        else:
            payload_dict["data"] = {"ap_results": response}
        return payload_dict

    def _plot_highlighted_overall_metrics(self, threshold: float) -> Dict[str, Any]:
        """
        Plot highlighted overall metrics at specified threshold.

        :param threshold: IoU threshold value
        :type threshold: float
        :return: Dictionary containing overall metrics data
        :rtype: Dict[str, Any]
        """
        rename_dict = {
            f"map{int(threshold*100)}": f"mAP@{int(threshold*100)}",
            f"map{int(threshold*100)}_11": f"mAP11@{int(threshold*100)}",
            "map50": "mAP@50",
            "map75": "mAP@75",
            "map5095": "mAP@[.50,.95]",
            "mar1": "mAR@max=1",
            "mar10": "mAR@max=10",
            "mar100": "mAR@max=100",
        }
        average_precision = AveragePrecision(
            class_mappings=self.class_mappings,
            coco_=self.coco_,
        )
        overall_metrics = average_precision.calculate_highlighted_overall_metrics(
            threshold
        )
        for metric in list(overall_metrics.keys()):
            overall_metrics[rename_dict[metric]] = overall_metrics.pop(metric)

        val_train_dict = {}
        for value in rename_dict.values():
            val_train_dict[value] = {"Validation": overall_metrics[value]}

        payload_dict = {"type": "overall", "data": val_train_dict}
        return payload_dict

    def _filter_ap_metrics(
        self, target_attribute_dict: Optional[Dict[str, Any]]
    ) -> List[int]:
        """
        Filter average precision metrics based on target attributes.

        :param target_attribute_dict: Dictionary of target attributes for filtering
        :type target_attribute_dict: Optional[Dict[str, Any]]
        :return: List of filtered indices
        :rtype: List[int]
        """
        if target_attribute_dict:
            idxs = self.validation_utils.filter_attribute_by_dict(
                target_attribute_dict
            ).index.tolist()

        return idxs

    def _plot_statistics_classbased_table(
        self,
        threshold: Optional[float] = None,
        target_attribute_dict: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Plot class-based statistics table.

        :param threshold: IoU threshold value
        :type threshold: Optional[float]
        :param target_attribute_dict: Dictionary for filtering by attributes
        :type target_attribute_dict: Optional[Dict[str, Any]]
        :return: Dictionary containing class-based statistics
        :rtype: Dict[str, Any]
        """

        average_precision = AveragePrecision(
            class_mappings=self.class_mappings,
            coco_=self.coco_,
        )

        rename_dict = {
            f"ap10": f"AP@10",
            f"ap10_11": f"AP11@10",
            "ap50": "AP@50",
            "ap75": "AP@75",
            "ap5095": "AP@[.50,.95]",
            "ar1": "AR@max=1",
            "ar10": "AR@max=10",
            "ar100": "AR@max=100",
        }
        if target_attribute_dict:
            idxs = self._filter_ap_metrics(target_attribute_dict)
        else:
            idxs = None

        ap_metrics = average_precision.calculate_ap_metrics(idxs=idxs)

        class_ap_metrics = dict()

        for class_ in ap_metrics["ap50"].keys():
            class_dict = dict()
            for metric in ap_metrics:
                class_dict[rename_dict[metric]] = ap_metrics[metric][class_]
            class_ap_metrics[self.class_mappings[str(class_)]] = class_dict

        payload_dict = {"type": "table", "data": {"Validation": class_ap_metrics}}
        return payload_dict

    def _plot_training_validation_comparison_classbased_table(self):

        threshold = 0.1
        payload_dict = self._plot_highlighted_overall_metrics(threshold)

        keys_to_be_included = ["mAP@50", "mAP@75", "mAP@[.50,.95]"]
        all_keys = payload_dict["data"].keys()

        for key in list(set(all_keys) - set(keys_to_be_included)):
            payload_dict["data"].pop(key)

        payload_dict["type"] = "bar"
        return payload_dict

    def _main_metric(self, threshold):
        average_precision = AveragePrecision(
            class_mappings=self.class_mappings,
            coco_=self.coco_,
        )
        mean_map_given = average_precision.calculate_ap_metrics(threshold=threshold)[
            "mAP"
        ].round(4)
        payload_dict = {f"mAP@{int(threshold*100)}": mean_map_given}
        return payload_dict

    def blind_spot_metrics(
        self, target_attribute_dict: Optional[Dict[str, Any]], threshold: float
    ) -> Dict[str, Any]:
        """
        Plots ROC Curve for target_class.
        References:
        https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html
        https://plotly.com/python/roc-and-pr-curves/
        :param target_class: target class to produce ROC plot
        :return: payload_dict
        """
        average_precision = AveragePrecision(
            class_mappings=self.class_mappings,
            coco_=self.coco_,
        )
        avg_rename_dict = {
            f"map{int(threshold*100)}": f"mAP@{int(threshold*100)}",
            f"map{int(threshold*100)}_11": f"mAP11@{int(threshold*100)}",
            "map50": "mAP@50",
            "map75": "mAP@75",
            "map5095": "mAP@[.50,.95]",
            "mar1": "mAR@max=1",
            "mar10": "mAR@max=10",
            "mar100": "mAR@max=100",
        }

        rename_dict = {
            "ap10": "AP@10",
            "ap10_11": "AP11@10",
            "ap50": "AP@50",
            "ap75": "AP@75",
            "ap5095": "AP@[.50,.95]",
            "ar1": "AR@max=1",
            "ar10": "AR@max=10",
            "ar100": "AR@max=100",
        }
        idxs = (
            self._filter_ap_metrics(target_attribute_dict)
            if bool(target_attribute_dict)
            else None
        )

        class_metrics = average_precision.calculate_ap_metrics(idxs=idxs)
        overall_metrics = average_precision.calculate_highlighted_overall_metrics(
            threshold
        )

        for key in list(class_metrics.keys()):
            class_metrics[rename_dict[key]] = class_metrics.pop(key)

        for metric in list(overall_metrics.keys()):
            overall_metrics[avg_rename_dict[metric]] = overall_metrics.pop(metric)

        blind_spot_metrics_dict = pd.DataFrame(class_metrics).T.to_dict()
        blind_spot_metrics_dict = {
            str(k): v for k, v in blind_spot_metrics_dict.items()
        }

        blind_spot_metrics_dict["Average"] = overall_metrics

        return blind_spot_metrics_dict



import bson
from typing import Union, Optional, List, Dict, Any, Callable


from gesund.core._exceptions import PlotError, MetricCalculationError
from gesund.core._schema import UserInputParams, UserInputData
from gesund.core._metrics.classification.classification_metric_plot import (
    Classification_Plot,
)
from gesund.core._metrics.object_detection.object_detection_metric_plot import (
    Object_Detection_Plot,
)
from gesund.core._metrics.semantic_segmentation.segmentation_metric_plot import (
    Semantic_Segmentation_Plot,
)
from gesund.core._data_loaders import DataLoader


class CommonPlots:
    def __init__(self):
        self.cls_driver = None
        self.obj_driver = None
        self.seg_driver = None

    def _class_distribution(
        self, metrics: Union[Dict, List], threshold: float, save_path: Optional[str]
    ):
        """
        A function to plot class distribution

        :param metrics: The metrics to be plotted
        :type metrics: Union[dict, list]
        :param threshold: the value to be applied as threshold
        :type threshold: float
        :param save_path: The path to be saved the plot in
        :type save_path: The path to be saved

        :return: None
        """
        self.cls_driver._plot_class_distributions(metrics, threshold, save_path)

    def _classification_blind_spot(
        self, metrics: Union[Dict, List], class_types: List, save_path: Optional[str]
    ):
        """
        A function to plot classification blind spot metrics

        :param metrics: The metrics to be plotted
        :type metrics: Union[dict, list]
        :param class_types: List of class types to include
        :type class_types: list
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):


        :return: None
        """
        self.cls_driver._plot_blind_spot(metrics, class_types, save_path)

    def _class_performance_by_threshold(
        self,
        metrics: Union[Dict, List],
        graph_type: str,
        threshold: float,
        save_path: Optional[str],
    ):
        """
        A function to plot class performance by threshold

        :param metrics: The metrics to be plotted
        :type metrics: Union[dict, list]
        :param graph_type: The type of graph to plot (e.g., 'graph_1')
        :type graph_type: str
        :param threshold: the value to be applied as threshold
        :type threshold: float
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):
        :return: None
        """
        self.cls_driver._plot_class_performance_by_threshold(
            graph_type, metrics, threshold, save_path
        )

    def _roc_statistics(self, roc_class: List, save_path: Optional[str]):
        """
        A function to plot ROC statistics

        :param metrics: The metrics to be plotted
        :type metrics: Union[dict, list]
        :param roc_class: List of classes to plot ROC curves for
        :type roc_class: List
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.cls_driver._plot_roc_statistics(roc_class, save_path)

    def _precision_recall_statistics(self, pr_class: List, save_path: Optional[str]):
        """
        A function to plot precision recall statistics

        :param pr_class: List of classes to plot precision recall curves for
        :type pr_class: List
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.cls_driver._plot_precision_recall_statistics(pr_class, save_path)

    def _classification_confidence_histogram(
        self, confidence_histogram_args: List, save_path: Optional[str]
    ):
        """
        A function to plot classification confidence histogram

        :param confidence_histogram_args: List of arguments for confidence histogram
        :type confidence_histogram_args: List
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.cls_driver._plot_confidence_histogram(confidence_histogram_args, save_path)

    def _classification_overall_metrics(
        self, overall_metrics_args: List, save_path: Optional[str]
    ):
        """
        A function to plot classification overall metrics

        :param overall_metrics_args: List of arguments for overall metrics
        :type overall_metrics_args: List
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.cls_driver._plot_overall_metrics(overall_metrics_args, save_path)

    def _confusion_matrix(self, save_path: Optional[str]):
        """
        A function to plot confusion matrix

        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.cls_driver._plot_confusion_matrix(save_path)

    def _prediction_dataset_distribution(self, save_path: Optional[str]):
        """
        A function to plot prediction dataset distribution

        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.cls_driver._plot_prediction_dataset_distribution(save_path)

    def _most_confused_bar(self, save_path: Optional[str]):
        """
        A function to plot most confused bar

        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.cls_driver._plot_most_confused_bar(save_path)

    def _confidence_histogram_scatter_distribution(self, save_path: Optional[str]):
        """
        A function to plot confidence histogram scatter distribution

        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.cls_driver._plot_confidence_histogram_scatter_distribution(save_path)

    def _lift_chart(self, save_path: Optional[str]):
        """
        A function to plot lift chart

        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.cls_driver._plot_lift_chart(save_path)

    def _object_detection_blind_spot(
        self, blind_spot_args: List, save_path: Optional[str]
    ):
        """
        A function to plot object detection blind spot

        :param blind_spot_args: List of arguments for blind spot
        :type blind_spot_args: List
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.obj_driver._plot_blind_spot(blind_spot_args, save_path)

    def _object_detection_overall_metrics(
        self, overall_metrics_args: List, save_path: Optional[str]
    ):
        """
        A function to plot object detection overall metrics

        :param overall_metrics_args: List of arguments for overall metrics
        :type overall_metrics_args: List
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.obj_driver._plot_overall_metrics(overall_metrics_args, save_path)

    def _top_misses(self, top_misses_args: List, save_path: Optional[str]):
        """
        A function to plot top misses

        :param top_misses_args: List of arguments for top misses
        :type top_misses_args: List
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.obj_driver._plot_top_misses(top_misses_args, save_path)

    def _classbased_table_metrics(
        self, classbased_table_args: List, save_path: Optional[str]
    ):
        """
        A function to plot class based table metrics

        :param classbased_table_args: List of arguments for class based table
        :type classbased_table_args: List
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.obj_driver._plot_classbased_table_metrics(classbased_table_args, save_path)

    def _mixed_metrics(self, mixed_metrics_args: List, save_path: Optional[str]):
        """
        A function to plot mixed metrics

        :param mixed_metrics_args: List of arguments for mixed metrics
        :type mixed_metrics_args: List
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.obj_driver._plot_mixed_metrics(mixed_metrics_args, save_path)

    def _object_detection_confidence_histogram(
        self, confidence_histogram_args: List, save_path: Optional[str]
    ):
        """
        A function to plot object detection confidence histogram

        :param confidence_histogram_args: List of arguments for confidence histogram
        :type confidence_histogram_args: List
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.obj_driver._plot_confidence_histogram(confidence_histogram_args, save_path)

    def _classbased_table(self, classbased_table_args: List, save_path: Optional[str]):
        """
        A function to plot class based table

        :param classbased_table_args: List of arguments for class based table
        :type classbased_table_args: List
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.seg_driver._plot_classbased_table(classbased_table_args, save_path)

    def _segmentation_overall_metrics(
        self, overall_args: List, save_path: Optional[str]
    ):
        """
        A function to plot overall data

        :param overall_args: List of arguments for overall data
        :type overall_args: List
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.seg_driver._plot_overall_data(overall_args, save_path)

    def _segmentation_blind_spot(self, blind_spot_args: List, save_path: Optional[str]):
        """
        A function to plot segmentation blind spot

        :param blind_spot_args: List of arguments for blind spot
        :type blind_spot_args: List
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.seg_driver._plot_blind_spot(blind_spot_args, save_path)

    def _by_meta_data(self, meta_data_args: List, save_path: Optional[str]):
        """
        A function to plot by meta data

        :param meta_data_args: List of arguments for meta data
        :type meta_data_args: List
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.seg_driver._plot_by_meta_data(meta_data_args, save_path)

    def _violin_graph(
        self, metrics: Union[Dict, List], threshold: float, save_path: Optional[str]
    ):
        """
        A function to plot violin graph

        :param metrics: The metrics to be plotted
        :type metrics: Union[Dict, List]
        :param threshold: the value to be applied as threshold
        :type threshold: float
        :param save_path: The path to be saved the plot in
        :type save_path: Optional[str]):

        :return: None
        """
        self.seg_driver._plot_violin_graph(metrics, threshold, save_path)


class ClassificationPlots(CommonPlots):
    def __init__(self):
        super().__init__()
        self.cls_driver = Classification_Plot()


class ObjectDetectionPlots(CommonPlots):
    def __init__(self):
        super().__init__()
        self.obj_driver = Object_Detection_Plot()


class SegmentationPlots(CommonPlots):
    def __init__(self):
        super().__init__()
        self.seg_driver = Semantic_Segmentation_Plot()


class PlotData:
    FXN_PLOT_MAP = {
        "classification": {
            "class_distribution": lambda self: self.classification_plotter._class_distribution,
            "blind_spot": lambda self: self.classification_plotter._classification_blind_spot,
            "class_performance_by_threshold": lambda self: self.classification_plotter._class_performance_by_threshold,
            "roc_statistics": lambda self: self.classification_plotter._roc_statistics,
            "precision_recall_statistics": lambda self: self.classification_plotter._precision_recall_statistics,
            "confidence_histogram": lambda self: self.classification_plotter._classification_confidence_histogram,
            "overall_metrics": lambda self: self.classification_plotter._classification_overall_metrics,
            "confusion_matrix": lambda self: self.classification_plotter._confusion_matrix,
            "prediction_dataset_distribution": lambda self: self.classification_plotter._prediction_dataset_distribution,
            "most_confused_bar": lambda self: self.classification_plotter._most_confused_bar,
            "confidence_histogram_scatter_distribution": lambda self: self.classification_plotter._confidence_histogram_scatter_distribution,
            "lift_chart": lambda self: self.classification_plotter._lift_chart,
        },
        "object_detection": {
            "mixed_metrics": lambda self: self.object_detection_plotter._mixed_metrics,
            "top_misses": lambda self: self.object_detection_plotter._top_misses,
            "confidence_histogram": lambda self: self.object_detection_plotter._object_detection_confidence_histogram,
            "classbased_table_metrics": lambda self: self.object_detection_plotter._classbased_table_metrics,
            "overall_metrics": lambda self: self.object_detection_plotter._object_detection_overall_metrics,
            "blind_spot": lambda self: self.object_detection_plotter._object_detection_blind_spot,
        },
        "segmentation": {
            "violin_graph": lambda self: self.segmentation_plotter._violin_graph,
            "by_meta_data": lambda self: self.segmentation_plotter._by_meta_data,
            "overall_metrics": lambda self: self.segmentation_plotter._segmentation_overall_metrics,
            "classbased_table": lambda self: self.segmentation_plotter._classbased_table,
            "blind_spot": lambda self: self.segmentation_plotter._segmentation_blind_spot,
        },
    }

    def __init__(
        self,
        metrics_result: Dict[str, Any],
        user_params: UserInputParams,
        user_data: Optional[UserInputData] = None,
        batch_job_id: Optional[str] = None,
        validation_problem_type_factory: Optional[Callable] = None,
    ):
        self.metrics_results = metrics_result
        self.user_params = user_params

        # set up data loading for plot if the plotData class is not part of the validation
        # execution workflow
        if not user_data:
            self.user_data = self._load_data()
        else:
            self.user_data = user_data

        # save the plots
        self.batch_job_id = batch_job_id
        if not batch_job_id:
            self.batch_job_id = str(bson.ObjectId())
        self.output_dir = f"outputs/{self.batch_job_id}"
        self.plot_save_dir = f"outputs/{self.batch_job_id}/plots"

        # instantiate the plotters
        self.classification_plotter = ClassificationPlots()
        self.object_detection_plotter = ObjectDetectionPlots()
        self.segmentation_plotter = SegmentationPlots()

        # import the validation problem type factory if plotData is not part of the validation
        # execution flow
        # if this class is part of the  validation execution flow, then importing the
        # validation problem type factory leads to circular imports
        if not validation_problem_type_factory:
            from gesund.validation import ValidationProblemTypeFactory

            self.validation_problem_type_factory = ValidationProblemTypeFactory()
        else:
            self.validation_problem_type_factory = validation_problem_type_factory

        # getter methods
        self.get_supported_plots = self.__get_supported_plots(user_params)

    def _load_data(self):
        """
        A function to load the JSON files

        :return: None
        """
        # Load data
        # set up source data for processing
        data_loader = DataLoader(self.user_params.data_format)
        data = {
            "prediction": data_loader.load(self.user_params.predictions_path),
            "annotation": data_loader.load(self.user_params.annotations_path),
        }
        if isinstance(self.user_params.class_mapping, str):
            data["class_mapping"] = data_loader.load(self.user_params.class_mapping)
        else:
            data["class_mapping"] = self.user_params.class_mapping

        # loading metadata
        if self.user_params.metadata_path:
            data["metadata"] = data_loader.load(self.user_params.metadata_path)

        # run conversion
        if self.user_params.json_structure_type != "gesund":
            self._convert_data(data)

        self.data = UserInputData(**data)

    @classmethod
    def __get_supported_plots(cls, user_params) -> List[str]:
        """
        Returns list of supported plots for the current problem type

        :param user_params: The parameters provided by the user
        :type user_params: UserInputParams

        :return: List of supported plot names
        :rtype: list
        """
        return list(cls.FXN_PLOT_MAP.get(user_params.problem_type, {}).keys())

    def _plot_single_metric(
        self, metric_name: str, threshold: float = 0.0, graph_type: str = "graph_1"
    ) -> None:
        """
        A function to return plotting function specific to metric

        :param metric_name: name of the metric
        :type metric_name: str
        :param threshold: threshold value for performance threshold plot
        :type threshold: float
        :param graph_type: type of graph for performance threshold plot
        :type graph_type: str

        :return: None
        """

        if self.user_params.problem_type not in self.FXN_PLOT_MAP:
            raise ModuleNotFoundError(
                f"Problem type {self.user_params.problem_type} not supported"
            )

        if metric_name not in self.FXN_PLOT_MAP[self.user_params.problem_type]:
            raise ModuleNotFoundError(f"{metric_name} plotting function not found")

        _plotter_fxn = self.FXN_PLOT_MAP[self.user_params.problem_type][metric_name](
            self
        )
        _plotter_fxn(
            metrics=self.metrics_results,
            threshold=threshold,
            save_path=self.plot_save_dir if self.user_params.save_plots else None,
        )

    def _plot_all(self, metric_validation_executor) -> None:
        """
        A function to plot all the metrics

        :param metric_validation_executor: metric validation executor
        :type metric_validation_executor: object

        :return: None
        """
        try:
            metric_validation_executor.plot_metrics(self.metrics_results)
        except Exception as e:
            print(e)
            raise MetricCalculationError("Could not run the validation executor !")

        if self.user_params.store_plots:
            pass

    def plot(self, metric_name: str = "all", threshold: float = 0.0) -> None:
        """
        Plot the data from the results

        :param metric_name: name of the metric to plot
        :type metric_name: str
        :param threshold: float value indicating the threshold
        :type threshold: float
        :return: None
        """
        if not self.batch_job_id:
            raise ValueError("batch_job_id is required for plotting")

        _validation_class = (
            self.validation_problem_type_factory.get_problem_type_factory(
                self.user_params.problem_type
            )
        )

        _metric_validation_executor = _validation_class(self.batch_job_id)

        # if the metadata path is provided then the metric result needs to be recalculated as per the
        # metadata filters of the interest
        if self.user_params.metadata_path:
            if self.user_data.was_converted:
                prediction = self.user_data.converted_prediction
                annotation = self.user_data.converted_annotation
            else:
                prediction = self.user_data.prediction
                annotation = self.user_data.annotation

            validation_data = (
                _metric_validation_executor.create_validation_collection_data(
                    prediction,
                    annotation,
                    self.user_data.metadata,
                )
            )
            metric_results = _metric_validation_executor.load(
                validation_data, self.user_data.class_mapping, self.user_data.metadata
            )
        else:
            metric_results = self.metrics_results

        if metric_name == "all":
            try:
                _metric_validation_executor.plot_metrics(
                    metric_results,
                    self.output_dir,
                    self.plot_save_dir,
                    self.user_params.plot_config,
                )
            except Exception as e:
                print(e)
                # raise PlotError("Could not plot the metrics!")
        else:
            self._plot_single_metric(metric_name, metric_results, threshold)



from collections import Counter
from typing import Dict, List, Optional, Union, Any
import numpy as np
import pandas as pd
from tqdm import tqdm

from gesund.core._utils import ValidationUtils, Statistics


class PlotObjectStats:
    def __init__(self, coco_, class_mappings, meta_data_dict=None):
        self.is_meta_exists = False
        self.coco_ = coco_

        if bool(meta_data_dict):
            self.is_meta_exists = True
            self.validation_utils = ValidationUtils(meta_data_dict)
        self.validation_utils = ValidationUtils(meta_data_dict)
        self.class_mappings = class_mappings
        self.class_idxs = [int(i) for i in list(class_mappings.keys())]

    def _plot_object_counts(
        self,
        confidence: float = 0,
        target_attribute_dict: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Plot object counts comparing ground truth and predictions.

        :param confidence: Confidence threshold for filtering predictions
        :type confidence: float
        :param target_attribute_dict: Dictionary for filtering by attributes
        :type target_attribute_dict: Optional[Dict[str, Any]]
        :return: Dictionary containing bar chart data of object counts
        :rtype: Dict[str, Any]
        """
        # Filter wrt target attribute dict
        idxs = self.validation_utils.filter_attribute_by_dict(
            target_attribute_dict
        ).index.tolist()

        # Counting the groundtruth
        gt_df = pd.DataFrame(self.coco_[1]["annotations"])
        gt_df = gt_df[gt_df["image_id"].isin(idxs)]

        gt_class_occurrence_df = (
            gt_df.groupby("image_id")["category_id"].value_counts().unstack().fillna(0)
        )
        gt_non_occurring_classes = list(
            set([int(i) for i in self.class_mappings.keys()])
            - set(gt_class_occurrence_df.columns.tolist())
        )
        for class_ in gt_non_occurring_classes:
            gt_class_occurrence_df[class_] = 0

        # Counting the predictions
        pred_df = pd.DataFrame(self.coco_[0])
        pred_df = pred_df[pred_df["score"] > confidence]
        pred_df = pred_df[pred_df["image_id"].isin(idxs)]

        pred_class_occurrence_df = (
            pred_df.groupby("image_id")["category_id"]
            .value_counts()
            .unstack()
            .fillna(0)
        )
        non_occurring_classes = list(
            set([int(i) for i in self.class_mappings.keys()])
            - set(pred_class_occurrence_df.columns.tolist())
        )
        for class_ in non_occurring_classes:
            pred_class_occurrence_df[class_] = 0

        # Sum Occurrences
        pred_count = pred_class_occurrence_df.sum().to_dict()
        gt_count = gt_class_occurrence_df.sum().to_dict()

        result_dict = dict()
        # Format them for frontend
        for cls_ in self.class_idxs:
            result_dict[self.class_mappings[str(cls_)]] = {
                "Predicted": pred_count[int(cls_)],
                "GT": gt_count[int(cls_)],
            }

        payload_dict = {
            "type": "bar",
            "data": result_dict,
        }
        return payload_dict

    def _plot_prediction_distribution(
        self, target_attribute_dict: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """
        Plot distribution of predictions across different classes.

        :param target_attribute_dict: Dictionary for filtering by attributes
        :type target_attribute_dict: Optional[Dict[str, Any]]
        :return: Dictionary containing pie chart data of prediction distribution
        :rtype: Dict[str, Any]
        """
        idxs = self.validation_utils.filter_attribute_by_dict(
            target_attribute_dict
        ).index.tolist()

        # Counting the predictions
        pred_df = pd.DataFrame(self.coco_[0])
        pred_df = pred_df[pred_df["image_id"].isin(idxs)]

        pred_class_occurrence_df = (
            pred_df.groupby("image_id")["category_id"]
            .value_counts()
            .unstack()
            .fillna(0)
        )
        non_occurring_classes = list(
            set([int(i) for i in self.class_mappings.keys()])
            - set(pred_class_occurrence_df.columns.tolist())
        )
        for class_ in non_occurring_classes:
            pred_class_occurrence_df[class_] = 0

        # Sum Occurrences
        pred_class_occurrence_df.columns = pred_class_occurrence_df.columns.astype(str)
        pred_class_occurrence_df.rename(columns=self.class_mappings, inplace=True)

        pred_class_occurrence_dict = pred_class_occurrence_df.sum().to_dict()

        payload_dict = {
            "type": "pie",
            "data": pred_class_occurrence_dict,
        }

        return payload_dict



class InputError(Exception):
    def __init__(self, msg: str):
        self.message = msg
        super().__init__(msg)

    def __str__(self):
        return self.message


class DataLoadError(Exception):
    def __init__(self, msg: str):
        self.message = msg
        super().__init__(msg)

    def __str__(self):
        return self.message


class MetricCalculationError(Exception):
    def __init__(self, msg: str):
        self.message = msg
        super().__init__(msg)

    def __str__(self):
        return self.message


class PlotError(Exception):
    def __init__(self, msg: str):
        self.message = msg
        super().__init__(msg)

    def __str__(self):
        return self.message



import numpy as np
import pandas as pd
import sklearn
from sklearn.metrics import auc
from sklearn.preprocessing import label_binarize
from sklearn.metrics import precision_recall_curve, roc_curve, average_precision_score

from .auc import AUC
from typing import Any, Dict, List, Optional, Union


class ThresholdMetrics:
    def __init__(
        self,
        true: Union[List[int], np.ndarray, pd.Series],
        pred_logits: pd.DataFrame,
        class_mappings: Dict[int, str],
    ) -> None:
        self.pred_logits = pred_logits
        self.true = true
        self.class_mappings = class_mappings

        self.auc = AUC(class_mappings=class_mappings)

    def calculate_statistics_per_class(
        self,
        true: Union[List[int], np.ndarray, pd.Series],
        pred_categorical: Union[List[int], np.ndarray, pd.Series],
    ) -> Dict[str, Any]:
        """
        Calculates class wise stats for the given class.
        :param target_class: Class to calculate metrics.
        :param true: True labels for samples
        :param pred_categorical: Prediction for samples
        :return: payload_dict
        """
        sense_spec_dict = self.auc.calculate_sense_spec(true, pred_categorical)

        F1 = sense_spec_dict["F1"][1]
        TPR = sense_spec_dict["TPR"][1]
        TNR = sense_spec_dict["TNR"][1]
        PPV = sense_spec_dict["PPV"][1]
        NPV = sense_spec_dict["NPV"][1]
        FPR = sense_spec_dict["FPR"][1]
        FNR = sense_spec_dict["FNR"][1]
        TP = sense_spec_dict["TP"][1]
        TN = sense_spec_dict["TN"][1]
        FP = sense_spec_dict["FP"][1]
        FN = sense_spec_dict["FN"][1]
        MCC = sense_spec_dict["Matthew's Classwise C C"][1]

        payload_dict = {
            "type": "bar",
            "data": {
                "graph_1": {
                    "F1": F1,
                    "Sensitivity": TPR,
                    "Specificity": TNR,
                    "Precision": PPV,
                    "Matthew's Classwise C C": MCC,
                    "FPR": FPR,
                    "FNR": FNR,
                },
                "graph_2": {"TP": TP, "TN": TN, "FP": FP, "FN": FN},
            },
        }
        return payload_dict



import numpy as np
import pandas as pd

from ..metrics.auc import AUC
from ..metrics.dataset_stats import DatasetStats
from gesund.core._utils import ValidationUtils, Statistics


class PlotPredictionDataAnalysis:
    def __init__(
        self,
        true,
        pred_logits,
        pred_categorical,
        meta_pred_true,
        class_mappings,
        meta,
    ):
        self.true = true
        self.pred_logits = pred_logits
        self.pred_categorical = pred_categorical
        self.class_mappings = class_mappings
        self.meta_pred_true = meta_pred_true
        self.meta = meta

        self.dataset_stats = DatasetStats()
        self.auc = AUC(class_mappings)
        self.validation_utils = ValidationUtils(meta_pred_true=meta_pred_true)

    def prediction_distribution(self, target_attribute_dict=None):
        """
        Calculates true positive, true negative, false positive, false negatives for the given class.
        :param target_class: Class to calculate metrics.
        :param true: True labels for samples
        :param pred_categorical: Prediction for samples
        :return: payload_dict
        """
        filtered_meta_pred_true = self.validation_utils.filter_attribute_by_dict(
            target_attribute_dict
        )
        pred_categorical = filtered_meta_pred_true["pred_categorical"]

        occurrences = pred_categorical.value_counts().sort_index()
        occurrences.index = occurrences.index.map(str)
        occurrences = occurrences.rename(index=self.class_mappings)
        occurrences = occurrences.to_dict()

        payload_dict = {
            "type": "square",
            "data": occurrences,
        }
        return payload_dict

    def gtless_confidence_histogram_scatter_distribution(
        self, predicted_class="overall", n_samples=300, randomize_x=True, n_bins=25
    ):
        # Plot Scatters

        # Filtering data
        filtered_pred_categorical = pd.DataFrame(self.pred_categorical.copy())
        if n_samples > filtered_pred_categorical.shape[0]:
            n_samples = filtered_pred_categorical.shape[0]
        filtered_pred_categorical = filtered_pred_categorical.sample(
            n_samples, replace=True
        )
        if predicted_class != "overall":
            filtered_pred_categorical = filtered_pred_categorical[
                filtered_pred_categorical["pred_categorical"] == predicted_class
            ]

        filtered_pred_logits = self.pred_logits[filtered_pred_categorical.index].max()
        filtered_pred_categorical["y"] = filtered_pred_logits

        # Renaming columns
        int_class_mappings = {int(k): v for k, v in self.class_mappings.items()}
        filtered_pred_categorical = filtered_pred_categorical.replace(
            {"pred_categorical": int_class_mappings}
        )
        filtered_pred_categorical = filtered_pred_categorical.rename(
            columns={"pred_categorical": "Prediction"}
        )

        # Randomize x if needed
        if randomize_x:
            filtered_pred_categorical["x"] = np.random.uniform(
                0, 1, filtered_pred_categorical.shape[0]
            )
        else:
            filtered_pred_categorical["x"] = filtered_pred_categorical["y"]

        points = list(
            filtered_pred_categorical.reset_index()
            .rename(columns={"index": "image_id"})
            .T.to_dict()
            .values()
        )

        # Plot histogram

        histogram = Statistics.calculate_histogram(
            filtered_pred_categorical["y"], min_=0, max_=1, n_bins=n_bins
        )

        payload_dict = {
            "type": "mixed",
            "data": {"points": points, "histogram": histogram},
        }

        return payload_dict

    def explore_predictions(self, predicted_class=None, top_k=3000):
        """
        Calculates minimum loss samples with appropriate information.
        :param self:
        :param predicted_class: Class of interest to calculate top losses.
        :param top_k: Number of most loss samples
        :return: payload_dict
        """

        # Check if overall top loss need to be observed.
        if predicted_class == "overall":
            predicted_class = None

        all_predictions = self.pred_categorical.copy(deep=True).head(top_k)
        predicted_logits = self.pred_logits.T.copy(deep=True)
        predicted_logits = predicted_logits.loc[all_predictions.index]
        meta = self.meta_pred_true[
            self.meta_pred_true.columns.difference(["pred_categorical", "true"])
        ]

        if predicted_class:
            logits_predicted_class_probability = predicted_logits[predicted_class]
        else:
            logits_predicted_class_probability = predicted_logits.max(axis=1)

        top_loss_data = []
        i = 1  # FIX.
        for idx in all_predictions.index:
            top_loss_data.append(
                {
                    "image_id": idx,
                    "rank": i,
                    "Prediction": self.class_mappings[
                        str(predicted_logits.loc[idx].idxmax())
                    ],
                    "Confidence": float(predicted_logits.loc[idx].max()),
                    "Meta Data": meta.loc[idx].to_dict(),
                }
            )
            i += 1
        payload_dict = {"type": "image", "data": top_loss_data}
        return payload_dict

    def softmax_probabilities_distribution(self, predicted_class, n_bins=25):
        payload_dict = {
            "type": "scatter",
            "data": Statistics.calculate_histogram(
                array_=self.pred_logits.T[predicted_class],
                min_=0,
                max_=1,
                n_bins=n_bins,
            ),
        }
        return payload_dict

    def class_distributions(self):
        """
        Calculates statistics on classes.
        :param true: true labels as a list = [1,0,3,4] for 4 sample dataset
        :param pred_categorical: categorical predictions as a list = [1,0,3,4] for 4 sample dataset
        :param labels: order of classes inside list
        :return: confusion matrix
        """
        data = {}
        counts = self.dataset_stats.calculate_class_distributions(self.true)
        validation_counts = counts[0]
        validation_counts_renamed = {
            self.class_mappings[str(k)]: v for k, v in validation_counts.items()
        }

        payload_dict = {
            "type": "bar",
            "data": {"Validation": validation_counts_renamed},
        }
        return payload_dict

    def meta_distributions(self):
        meta_counts = self.dataset_stats.calculate_meta_distributions(self.meta)
        data_dict = {}
        data_dict["Validation"] = meta_counts
        payload_dict = {}
        payload_dict["type"] = "bar"
        payload_dict["data"] = data_dict
        return payload_dict

    def prediction_dataset_distribution(self, target_attribute_dict=None):
        filtered_meta_pred_true = self.validation_utils.filter_attribute_by_dict(
            target_attribute_dict
        )

        # Dataset Distribution
        true_ = filtered_meta_pred_true["true"]
        data = {}
        dataset_class_counts = self.dataset_stats.calculate_class_distributions(true_)
        dataset_class_counts = dataset_class_counts[0]
        dataset_class_counts = {
            self.class_mappings[str(k)]: v for k, v in dataset_class_counts.items()
        }

        # Prediction Distribution
        pred_categorical = filtered_meta_pred_true["pred_categorical"]

        pred_class_counts = pred_categorical.value_counts().sort_index()
        pred_class_counts.index = pred_class_counts.index.map(str)
        pred_class_counts = pred_class_counts.rename(index=self.class_mappings)
        pred_class_counts = pred_class_counts.to_dict()

        payload_dict = {}
        payload_dict["type"] = "square"
        payload_dict["data"] = {
            "Annotation": dataset_class_counts,
            "Prediction": pred_class_counts,
        }
        return payload_dict



import numpy as np
import pandas as pd
from sklearn.metrics import auc
import sklearn
from typing import Any, Dict, List, Optional, Union


class LiftChart:
    def __init__(self, class_mappings: Dict[int, str]) -> None:
        self.class_mappings = class_mappings
        self.class_order = [int(i) for i in list(class_mappings.keys())]

    def _decile_table(
        self,
        true: Union[List[int], np.ndarray, pd.Series],
        pred_logits: pd.DataFrame,
        predicted_class: int = 0,
        change_deciles: int = 20,
        labels: bool = True,
        round_decimal: int = 3,
    ) -> pd.DataFrame:
        """Generates the Decile Table from labels and probabilities

        The Decile Table is creared by first sorting the customers by their predicted
        probabilities, in decreasing order from highest (closest to one) to
        lowest (closest to zero). Splitting the customers into equally sized segments,
        we create groups containing the same numbers of customers, for example, 10 decile
        groups each containing 10% of the customer base.
        """
        df = pd.DataFrame([true, pred_logits.loc[predicted_class].T]).T
        df = df.rename(columns={predicted_class: "pred_logits"})

        # df['decile']=pd.qcut(df['pred_logits'], 10, labels=list(np.arange(10,0,-1)))
        # ValueError: Bin edges must be unique

        change_deciles = max(change_deciles, len(true))

        df.sort_values("pred_logits", ascending=False, inplace=True)
        df["decile"] = np.linspace(1, change_deciles + 1, len(df), False, dtype=int)

        # lift_df abbreviation for decile_table
        lift_df = (
            df.groupby("decile")
            .apply(
                lambda x: pd.Series(
                    [
                        np.min(x["pred_logits"]),
                        np.max(x["pred_logits"]),
                        np.mean(x["pred_logits"]),
                        np.size(x["pred_logits"]),
                        np.sum(x["true"]),
                        np.size(x["true"][x["true"] == 0]),
                    ],
                    index=(
                        [
                            "prob_min",
                            "prob_max",
                            "prob_avg",
                            "cnt_cust",
                            "cnt_resp",
                            "cnt_non_resp",
                        ]
                    ),
                )
            )
            .reset_index()
        )

        lift_df["prob_min"] = lift_df["prob_min"].round(round_decimal)
        lift_df["prob_max"] = lift_df["prob_max"].round(round_decimal)
        lift_df["prob_avg"] = round(lift_df["prob_avg"], round_decimal)

        tmp = df[["true"]].sort_values("true", ascending=False)
        tmp["decile"] = np.linspace(1, change_deciles + 1, len(tmp), False, dtype=int)

        lift_df["cnt_resp_rndm"] = np.sum(df["true"]) / change_deciles
        lift_df["cnt_resp_wiz"] = tmp.groupby("decile", as_index=False)["true"].sum()[
            "true"
        ]

        lift_df["resp_rate"] = round(
            lift_df["cnt_resp"] * 100 / lift_df["cnt_cust"], round_decimal
        )
        lift_df["cum_cust"] = np.cumsum(lift_df["cnt_cust"])
        lift_df["cum_resp"] = np.cumsum(lift_df["cnt_resp"])
        lift_df["cum_resp_wiz"] = np.cumsum(lift_df["cnt_resp_wiz"])
        lift_df["cum_non_resp"] = np.cumsum(lift_df["cnt_non_resp"])
        lift_df["cum_cust_pct"] = round(
            lift_df["cum_cust"] * 100 / np.sum(lift_df["cnt_cust"]), round_decimal
        )
        lift_df["cum_resp_pct"] = round(
            lift_df["cum_resp"] * 100 / np.sum(lift_df["cnt_resp"]), round_decimal
        )
        lift_df["cum_resp_pct_wiz"] = round(
            lift_df["cum_resp_wiz"] * 100 / np.sum(lift_df["cnt_resp_wiz"]),
            round_decimal,
        )
        lift_df["cum_non_resp_pct"] = round(
            lift_df["cum_non_resp"] * 100 / np.sum(lift_df["cnt_non_resp"]),
            round_decimal,
        )
        lift_df["KS"] = round(
            lift_df["cum_resp_pct"] - lift_df["cum_non_resp_pct"], round_decimal
        )
        lift_df["lift"] = round(
            lift_df["cum_resp_pct"] / lift_df["cum_cust_pct"], round_decimal
        )
        return lift_df

    def calculate_lift_curve_points(
        self,
        true: Union[List[int], np.ndarray, pd.Series],
        pred_logits: pd.DataFrame,
        predicted_class: Optional[int] = None,
        change_deciles: int = 20,
        labels: bool = True,
        round_decimal: int = 3,
    ) -> Dict[str, List[Dict[str, float]]]:
        class_lift_dict = dict()
        if predicted_class in [None, "all", "overall"]:
            for class_ in list(self.class_mappings.keys()):
                lift_df = self._decile_table(
                    true,
                    pred_logits,
                    predicted_class=int(class_),
                    change_deciles=20,
                    labels=True,
                    round_decimal=3,
                )
                xy_points = [
                    {"x": i[0], "y": i[1]}
                    for i in lift_df[["prob_avg", "lift"]].values.tolist()
                ]
                class_lift_dict[self.class_mappings[str(class_)]] = xy_points
        else:
            lift_df = self._decile_table(
                true,
                pred_logits,
                predicted_class=predicted_class,
                change_deciles=20,
                labels=True,
                round_decimal=3,
            )
            xy_points = [
                {"x": i[0], "y": i[1]}
                for i in lift_df[["prob_avg", "lift"]].values.tolist()
            ]
            class_lift_dict[self.class_mappings[str(predicted_class)]] = xy_points

        return class_lift_dict



import os
import warnings
from tqdm import tqdm
import pandas as pd
from typing import Dict, List, Optional, Union, Any

from .average_precision import PlotAveragePrecision
from .top_losses import PlotTopLosses
from .object_stats import PlotObjectStats
from .confidence_distribution import PlotConfidenceGraphs
from .dataset_stats import PlotDatasetStats


class ObjectDetectionPlotDriver:
    def __init__(
        self,
        coco_,
        class_mappings,
        ground_truth_dict,
        prediction_dict,
        batch_job_id,
        meta_data_dict=None,
        loss_dict=None,
        filtering_meta=None,
    ):
        # Create validation variables
        self.true = ground_truth_dict
        self.pred = prediction_dict
        self.class_mappings = class_mappings
        self.meta = None
        self.batch_job_id = batch_job_id
        self.coco_ = coco_
        self.filtering_meta = filtering_meta
        if meta_data_dict:
            self.meta = pd.DataFrame(meta_data_dict).T
        if loss_dict:
            self.loss = pd.DataFrame(loss_dict, index=[0])

        self.sample_size = len(self.true)
        self.class_order = list(range(len(class_mappings.keys())))
        # Check, add breakpoint

        # Import Classes
        self.plot_average_precision = PlotAveragePrecision(
            class_mappings=self.class_mappings,
            meta_data_dict=meta_data_dict,
            coco_=coco_,
        )

        self.plot_object_stats = PlotObjectStats(
            coco_=coco_,
            class_mappings=self.class_mappings,
            meta_data_dict=meta_data_dict,
        )

        self.plot_confidence_graphs = PlotConfidenceGraphs(
            class_mappings=self.class_mappings,
            meta_data_dict=meta_data_dict,
            coco_=coco_,
        )

        self.plot_dataset_stats = PlotDatasetStats(
            class_mappings=self.class_mappings,
            meta_data_dict=meta_data_dict,
        )

        self.plot_loss = PlotTopLosses(
            coco_=coco_,
            class_mappings=self.class_mappings,
            meta_dict=self.meta,
        )

    def plot_highlighted_overall_metrics(self) -> Dict[str, Any]:
        """
        Plot highlighted overall metrics at IoU threshold 0.1.

        :return: Dictionary containing overall metrics data
        :rtype: Dict[str, Any]
        """
        return self.plot_average_precision._plot_highlighted_overall_metrics(
            threshold=0.1
        )

    def plot_performance_by_iou_threshold(
        self, threshold: float = 0.5, return_points: bool = False
    ) -> Dict[str, Any]:
        """
        Plot performance metrics at specified IoU threshold.

        :param threshold: IoU threshold value
        :type threshold: float
        :param return_points: Whether to return coordinate points
        :type return_points: bool
        :return: Dictionary containing performance metrics
        :rtype: Dict[str, Any]
        """
        return self.plot_average_precision._plot_performance_by_iou_threshold(
            threshold, return_points
        )

    def plot_statistics_classbased_table(
        self, target_attribute_dict: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """
        Plot statistics table based on classes.

        :param target_attribute_dict: Optional dictionary for filtering attributes
        :type target_attribute_dict: Optional[Dict[str, Any]]
        :return: Dictionary containing class-based statistics
        :rtype: Dict[str, Any]
        """
        return self.plot_average_precision._plot_statistics_classbased_table(
            threshold=0.1, target_attribute_dict=self.filtering_meta
        )

    def plot_object_counts(
        self,
        confidence: float = 0,
        target_attribute_dict: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Plot object count distributions.

        :param confidence: Confidence threshold for filtering
        :type confidence: float
        :param target_attribute_dict: Optional dictionary for filtering attributes
        :type target_attribute_dict: Optional[Dict[str, Any]]
        :return: Dictionary containing object count data
        :rtype: Dict[str, Any]
        """
        return self.plot_object_stats._plot_object_counts(
            confidence=confidence, target_attribute_dict=self.filtering_meta
        )

    def plot_top_misses(self, top_k: int = 100) -> Dict[str, Any]:
        """
        Plot top misclassified samples.

        :param top_k: Number of top misses to plot
        :type top_k: int
        :return: Dictionary containing top misses data
        :rtype: Dict[str, Any]
        """
        return self.plot_loss._plot_top_misses(top_k=top_k)

    def plot_confidence_histogram_scatter_distribution(
        self, predicted_class: Optional[str] = None, n_samples: int = 300
    ) -> Dict[str, Any]:
        """
        Plot confidence distribution as histogram and scatter plot.

        :param predicted_class: Class name to filter results
        :type predicted_class: Optional[str]
        :param n_samples: Number of samples to plot
        :type n_samples: int
        :return: Dictionary containing distribution plots data
        :rtype: Dict[str, Any]
        """
        return (
            self.plot_confidence_graphs._plot_confidence_histogram_scatter_distribution(
                predicted_class, n_samples
            )
        )

    def plot_prediction_distribution(
        self, target_attribute_dict: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """
        Plot distribution of predictions.

        :param target_attribute_dict: Optional dictionary for filtering attributes
        :type target_attribute_dict: Optional[Dict[str, Any]]
        :return: Dictionary containing prediction distribution data
        :rtype: Dict[str, Any]
        """
        return self.plot_object_stats._plot_prediction_distribution(
            target_attribute_dict=self.filtering_meta
        )

    def plot_meta_distribution(self) -> Dict[str, Any]:
        """
        Plot distribution of metadata attributes.

        :return: Dictionary containing metadata distribution plots
        :rtype: Dict[str, Any]
        """
        return self.plot_dataset_stats._plot_meta_distributions()

    def plot_training_validation_comparison_classbased_table(self) -> Dict[str, Any]:
        """
        Plot comparison table between training and validation results.

        :return: Dictionary containing comparison table data
        :rtype: Dict[str, Any]
        """
        return (
            self.plot_average_precision._plot_training_validation_comparison_classbased_table()
        )

    def main_metric(self) -> Dict[str, float]:
        """
        Calculate main evaluation metric.

        :return: Dictionary containing main metric value
        :rtype: Dict[str, float]
        """
        return self.plot_average_precision._main_metric(threshold=0.1)

    # Blind Spots
    def plot_blind_spot_metrics(
        self, target_attribute_dict: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """
        Plot metrics for model blind spots.

        :param target_attribute_dict: Optional dictionary for filtering attributes
        :type target_attribute_dict: Optional[Dict[str, Any]]
        :return: Dictionary containing blind spot metrics
        :rtype: Dict[str, Any]
        """
        return self.plot_average_precision.blind_spot_metrics(
            target_attribute_dict=self.filtering_meta, threshold=0.1
        )

    def _calling_all_plots(self) -> Dict[str, Any]:
        """
        Execute all plotting methods in the class.

        :return: Dictionary containing results from all plotting methods
        :rtype: Dict[str, Any]
        """
        # Getting all methods that do not start with '_'
        methods = [
            method_name
            for method_name in dir(self)
            if callable(getattr(self, method_name)) and not method_name.startswith("_")
        ]
        results = {}

        for method_name in tqdm(methods, desc="Calling all plot functions"):
            method = getattr(self, method_name)
            try:
                # Suppress warnings
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")

                    print(f"Calling method: {method_name}...")
                    # Attempt to call the method, handle cases where no arguments are required
                    result = method()
                    results[method_name] = result
            except TypeError as e:
                results[method_name] = f"Could not call {method_name}: {str(e)}"
            except Exception as e:
                results[method_name] = f"Error in {method_name}: {str(e)}"

        return results



import numpy as np
import pandas as pd
from sklearn.metrics import auc
from typing import Any, Dict, List, Optional, Union


class Accuracy:
    def __init__(self, class_mappings: Optional[Dict[int, str]] = None) -> None:
        self.class_mappings = class_mappings
        if self.class_mappings:
            self.class_order = [int(i) for i in list(class_mappings.keys())]

    def _calculate_accuracy(self, true, pred_categorical):
        if len(true) != 0:
            return float(np.sum(true == pred_categorical) / len(true))
        else:
            return 0

    def calculate_accuracy(
        self,
        true: Union[np.ndarray, pd.Series],
        pred_categorical: Union[np.ndarray, pd.Series],
        target_class: str = None,
    ) -> float:
        """
        Returns accuracy for all dataset between 0-1.
        return: accuracy between 0-1.
        """
        if true is None:
            true = self.true
        if pred_categorical is None:
            pred_categorical = self.pred_categorical

        if target_class == "overall" or target_class is None:
            return self._calculate_accuracy(true, pred_categorical)

        elif target_class == "all":
            class_accuracies = {}
            for target_class in self.class_order:
                true_idx = true == target_class
                class_true = true[true_idx]
                class_pred_categorical = pred_categorical[true_idx]
                class_accuracies[target_class] = self._calculate_accuracy(
                    class_true, class_pred_categorical
                )
            return class_accuracies

        elif target_class is not None:
            true_idx = true == target_class
            class_true = true[true_idx]
            class_pred_categorical = pred_categorical[true_idx]
            return self._calculate_accuracy(class_true, class_pred_categorical)
        else:
            return self._calculate_accuracy(true, pred_categorical)



import itertools
from datetime import datetime
import os
import pickle

import numpy as np
import pandas as pd
import sklearn
from pycocotools.coco import COCO
from pycocotools.cocoeval import COCOeval


class AveragePrecision:
    def __init__(self, class_mappings, coco_):
        self.class_mappings = class_mappings
        self.class_idxs = [int(i) for i in list(class_mappings.keys())]
        self.coco_ = coco_

    def calculate_coco_metrics(
        self,
        pred_coco,
        gt_coco,
        return_class_metrics=False,
        return_points=False,
        return_conf_dist=False,
        threshold=None,
        top_losses=False,
        idxs=None,
    ):
        annType = ["segm", "bbox", "keypoints"]
        annType = annType[1]  # specify type here

        annFile = gt_coco
        cocoGt = COCO(annFile)

        resFile = pred_coco
        cocoDt = cocoGt.loadRes(resFile)

        imgIds = sorted(cocoGt.getImgIds())
        cocoEval = COCOeval(cocoGt, cocoDt, annType)

        if top_losses:
            class_mean_list = []
            losses_list = []
            ids = cocoEval.params.catIds
            cocoEval.params.imgIds = imgIds
            cocoEval.evaluate()
            cocoEval.accumulate()
            for img in imgIds:
                for ids_ in ids:
                    a = cocoEval.ious[(img, ids_)]
                    if len(a) != 0:
                        class_mean = a.max(1).mean()
                        class_mean_list.append(class_mean)
                    else:
                        pass

                mean = sum(class_mean_list) / len(class_mean_list)
                losses_list.append({"image_id": img, "mIoU": mean})

            sorted_list = sorted(losses_list, key=lambda x: x["mIoU"], reverse=True)
            for i, item in enumerate(sorted_list):
                item["rank"] = i + 1
            return losses_list

        if return_conf_dist:
            cocoEval.params.imgIds = imgIds
            cocoEval.evaluate()
            cocoEval.accumulate()
            eval_imgs = [i for i in cocoEval.evalImgs if i is not None]

            return eval_imgs

        if return_points:
            xy_list = {}
            ids = cocoEval.params.catIds
            cocoEval.params.imgIds = imgIds
            cocoEval.evaluate()
            cocoEval.accumulate()

            for ids_ in ids:
                coordinates = cocoEval.plotPrecisionRecallGraph(threshold, ids_)
                xy_list[ids_] = coordinates

            return xy_list

        if return_class_metrics:

            metrics = []
            ids = cocoEval.params.catIds

            if idxs:
                cocoEval.params.imgIds = idxs

            for ids_ in ids:

                cocoEval.params.catIds = [ids_]  # Class-wise metrics outputted
                cocoEval.evaluate()  # [ 0_dict(), 1_dict(), 2_dict(), ...]
                cocoEval.accumulate()
                if threshold:
                    metric = cocoEval.summarize(threshold)
                else:
                    metric = cocoEval.summarize()

                metrics.append(metric)

            metrics_by_class = {
                metric: {i: metrics[i][metric] for i in range(len(ids))}
                for metric in metrics[0]
            }
            metrics_final = {k.replace("m", ""): v for k, v in metrics_by_class.items()}
            if threshold:
                metrics_final["APs"] = metrics_final.pop(f"ap{threshold}")
                metrics_final["mAP"] = np.mean(list(metrics_final["APs"].values()))

        else:
            cocoEval.params.imgIds = imgIds
            cocoEval.evaluate()
            cocoEval.accumulate()
            metrics_final = cocoEval.summarize()

        return metrics_final

    def calculate_highlighted_overall_metrics(self, threshold):

        pred_coco = self.coco_[0]
        gt_coco = self.coco_[1]

        return self.calculate_coco_metrics(pred_coco, gt_coco)

    def calculate_ap_metrics(
        self, threshold=None, idxs=None, return_map=False, return_class_metrics=True
    ):
        pred_coco = self.coco_[0]
        gt_coco = self.coco_[1]

        class_based_coco_metrics = self.calculate_coco_metrics(
            pred_coco,
            gt_coco,
            return_class_metrics=return_class_metrics,
            threshold=threshold,
            idxs=idxs,
        )

        return class_based_coco_metrics

    def calculate_iou_threshold_graph(self, threshold, return_points=True):

        pred_coco = self.coco_[0]
        gt_coco = self.coco_[1]

        return self.calculate_coco_metrics(
            pred_coco, gt_coco, return_points=return_points, threshold=threshold
        )

    def calculate_confidence_distribution(
        self, threshold, predicted_class=None, return_conf_dist=True
    ):
        pred_coco = self.coco_[0]
        gt_coco = self.coco_[1]
        eval_imgs = self.calculate_coco_metrics(
            pred_coco, gt_coco, return_conf_dist=return_conf_dist
        )
        existing_image_ids = [i["image_id"] for i in eval_imgs]
        image_id_scores = dict()
        for image_id in existing_image_ids:
            for cls_id in self.class_mappings:
                cls_id = int(cls_id)
                if predicted_class is not None and cls_id != predicted_class:
                    continue
                confidences = [
                    i["dtScores"]
                    for i in eval_imgs
                    if i["image_id"] == image_id and i["category_id"] == cls_id
                ]
                if len(confidences) != 0:
                    image_id_scores[image_id] = np.mean(np.array(confidences).flatten())
                    break
        return image_id_scores

    def plot_top_losses(self, top_losses=True):

        pred_coco = self.coco_[0]
        gt_coco = self.coco_[1]

        response = self.calculate_coco_metrics(
            pred_coco, gt_coco, top_losses=top_losses
        )

        return response



import numpy as np
import pandas as pd

from ..metrics.average_precision import AveragePrecision


class TopLosses:
    def __init__(self, coco_, class_mappings, loss):
        self.coco_ = coco_
        self.class_mappings = class_mappings

    def calculate_top_losses(self, top_k=100):
        """
        Sorts top loss samples given the target class.
        :param self:
        :param target_class:
        :param top_k: top_k top loss sample to show
        :return: sorted top loss
        """
        average_precision = AveragePrecision(
            class_mappings=self.class_mappings,
            coco_=self.coco_,
        )

        top_losses_list = average_precision.plot_top_losses()

        return top_losses_list



import json
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import os


class Semantic_Segmentation_Plot:
    """
    A class for creating various plots and visualizations for semantic segmentation results.

    :param violin_path: (str, optional) Path to the violin plot JSON data file.
    :param result_dict: (dict, optional) Dictionary containing result data.
    :param classed_table: (str, optional) Path to the class-based table JSON data file.
    :param overall_data: (str, optional) Path to the overall metrics JSON data file.
    :param blind_spot: (str, optional) Path to the blind spot analysis JSON data file.
    :param plot_by_meta_data: (str, optional) Path to the metadata plot JSON data file.
    :param output_dir: (str, optional) Directory path for saving output plots.
    """

    def __init__(
        self,
        violin_path=None,
        result_dict=None,
        classed_table=None,
        overall_data=None,
        blind_spot=None,
        plot_by_meta_data=None,
        output_dir=None,
    ):
        self.output_dir = output_dir
        if violin_path:
            self.violin_data = self._load_json(violin_path)
        if result_dict:
            self.result_dict = result_dict
        if classed_table:
            self.classbased_table = self._load_json(classed_table)
        if overall_data:
            self.overall_data = self._load_json(overall_data)
        if blind_spot:
            self.blind_spot_data = self._load_json(blind_spot)
        if plot_by_meta_data:
            self.plot_by_meta_data = self._load_json(plot_by_meta_data)

    def _load_json(self, json_path):
        """
        Load data from a JSON file.

        :param json_path: (str) Path to the JSON file to be loaded.
        :return: (dict) Loaded JSON data as a Python dictionary.
        """
        with open(json_path, "r") as file:
            data = json.load(file)
        return data

    def draw(
        self,
        plot_type,
        metrics=None,
        threshold=None,
        classbased_table_args=None,
        overall_args=None,
        blind_spot_args=None,
        meta_data_args=None,
        save_path=None,
    ):
        """
        Draw various types of plots based on the specified plot type and parameters.

        :param plot_type: (str) Type of plot to generate ('violin_graph', 'classbased_table', 'overall_metrics', 'blind_spot', 'plot_by_meta_data').
        :param metrics: (list, optional) List of metrics to plot for violin graph.
        :param threshold: (float, optional) Threshold value for violin graph.
        :param classbased_table_args: (dict, optional) Arguments for class-based table plot.
        :param overall_args: (dict, optional) Arguments for overall metrics plot.
        :param blind_spot_args: (dict, optional) Arguments for blind spot analysis plot.
        :param meta_data_args: (dict, optional) Arguments for metadata-based plot.
        :param save_path: (str, optional) Path to save the generated plot.

        :raises ValueError: If an unsupported plot type is specified.
        """
        if plot_type == "violin_graph":
            self._plot_violin_graph(metrics, threshold, save_path)
        elif plot_type == "classbased_table":
            self._plot_classbased_table(classbased_table_args, save_path)
        elif plot_type == "overall_metrics":
            self._plot_overall_data(overall_args, save_path)
        elif plot_type == "blind_spot":
            self._plot_blind_spot(blind_spot_args, save_path)
        elif plot_type == "plot_by_meta_data":
            self._plot_by_meta_data(meta_data_args, save_path)
        else:
            raise ValueError(f"Unsupported plot type: {plot_type}")

    def _plot_violin_graph(self, metrics=None, threshold=None, save_path=None):
        """
        Create a violin plot visualization for specified metrics data.

        :param metrics: (list, optional) List of metric names to include in the plot. If None, all metrics will be plotted.
        :param threshold: (float, optional) Minimum threshold value for filtering data points.
        :param save_path: (str, optional) Path where the plot should be saved. If None, saves as 'violin_plot.png'.

        :return: None
        :raises AttributeError: If no valid violin data is loaded.
        """
        if not hasattr(self, "violin_data") or self.violin_data.get("type") != "violin":
            print("No valid 'violin' data found in the JSON.")
            return

        data = self.violin_data.get("data", {})
        df = pd.DataFrame(data)

        if metrics:
            df = df[metrics]

        if threshold is not None:
            df = df[df.apply(lambda x: x > threshold)]

        plt.figure(figsize=(14, 8))
        sns.violinplot(data=df, palette="pastel", linewidth=1.5, inner="box")
        plt.title("Violin Plot of Metrics", fontsize=20, fontweight="bold", pad=20)

        plt.xlabel("Metrics", fontsize=14, labelpad=15)
        plt.ylabel("Values", fontsize=14, labelpad=15)

        plt.xticks(fontsize=12)
        plt.yticks(fontsize=12)

        plt.grid(True, axis="y", linestyle="--", alpha=0.7)

        if save_path:
            plt.savefig(save_path, bbox_inches="tight", dpi=300)
        else:
            plt.savefig("violin_plot.png", bbox_inches="tight", dpi=300)

        plt.show()
        plt.close()

    def _plot_classbased_table(self, classbased_table_args=None, save_path=None):
        """
        Create a heatmap visualization of class-based metrics data.

        :param classbased_table_args: (dict, optional) Dictionary containing plotting arguments.
                                    Expected format: {'classbased_table_args': threshold_value}
        :param save_path: (str, optional) Path where the plot should be saved. If None, saves as 'classbased_table_plot.png'.

        :return: None
        :raises AttributeError: If no valid table data is loaded.
        :note: The heatmap uses a coolwarm color scheme and displays values with 3 decimal places.
        """
        if (
            not hasattr(self, "classbased_table")
            or self.classbased_table.get("type") != "table"
        ):
            print("No valid 'table' data found in the JSON.")
            return

        data = self.classbased_table.get("data", {})
        flattened_data = {}

        for category, metrics in data.items():
            for subcategory, values in metrics.items():
                flattened_data[f"{category}_{subcategory}"] = values

        df = pd.DataFrame(flattened_data).T

        threshold = 0.0
        if (
            isinstance(classbased_table_args, dict)
            and "classbased_table_args" in classbased_table_args
        ):
            threshold = classbased_table_args["classbased_table_args"]

        df = df[df > threshold]
        df = df.replace(0, np.nan)

        df = df.dropna(how="all").dropna(axis=1, how="all")

        plt.figure(figsize=(16, 10))

        sns.heatmap(
            df,
            annot=True,
            cmap="coolwarm",
            fmt=".3f",
            cbar=True,
            linewidths=0.5,
            linecolor="gray",
        )

        plt.title(
            "Class-based Table Plot of Metrics", fontsize=20, fontweight="bold", pad=20
        )
        plt.xlabel("Metrics", fontsize=14, labelpad=15)
        plt.ylabel("Categories", fontsize=14, labelpad=15)

        plt.tight_layout()

        if save_path:
            plt.savefig(save_path, bbox_inches="tight", dpi=300)
        else:
            plt.savefig("classbased_table_plot.png", bbox_inches="tight", dpi=300)

        plt.show()
        plt.close()

    def _plot_overall_data(self, overall_args=None, save_path=None):
        """
        Create a bar plot visualization for overall validation metrics.

        :param overall_args: (list, optional) List of specific metrics to plot. If None, all metrics will be plotted.
        :param save_path: (str, optional) Path where the plot should be saved. If None, saves as 'overall_metrics_plot.png'.

        :return: None
        :raises AttributeError: If no valid overall data is loaded.
        """
        if (
            not hasattr(self, "overall_data")
            or self.overall_data.get("type") != "overall"
        ):
            print("No valid 'overall' data found in the JSON.")
            return

        data = self.overall_data.get("data", {})
        df = pd.DataFrame({k: v["Validation"] for k, v in data.items()}, index=[0])

        if overall_args:
            df = df[overall_args]

        df = df.T.reset_index()
        df.columns = ["Metric", "Value"]

        plt.figure(figsize=(14, 8))
        sns.barplot(
            x="Metric",
            y="Value",
            hue="Metric",
            data=df,
            palette="pastel",
            edgecolor="black",
            legend=False,
        )
        plt.title("Overall Metrics", fontsize=20, fontweight="bold", pad=20)
        plt.xlabel("Metrics", fontsize=14, labelpad=15)
        plt.ylabel("Values", fontsize=14, labelpad=15)

        plt.xticks(rotation=45, ha="right", fontsize=12)
        plt.grid(True, axis="y", linestyle="--", alpha=0.7)

        for i, row in enumerate(df.itertuples()):
            plt.text(
                i,
                row.Value + 0.02,
                f"{row.Value:.4f}",
                ha="center",
                va="bottom",
                fontsize=12,
                color="black",
                fontweight="bold",
            )

        plt.tight_layout()

        if save_path:
            plt.savefig(save_path, bbox_inches="tight", dpi=300)
        else:
            plt.savefig("overall_metrics_plot.png", bbox_inches="tight", dpi=300)

        plt.show()
        plt.close()

    def _plot_blind_spot(self, blind_spot_args=None, save_path=None):
        """
        Create a bar plot visualization for blind spot metrics comparison.

        :param blind_spot_args: (list, optional) List of specific metrics to plot. Invalid metrics will be filtered out.
        :param save_path: (str, optional) Path where the plot should be saved. If None, saves as 'blind_spot_metrics_plot.png'.

        :return: None
        :raises AttributeError: If no valid blind spot data is loaded.
        """

        if not hasattr(self, "blind_spot_data"):
            print("No valid 'blind_spot' data found in the JSON.")
            return

        if "Average" in self.blind_spot_data:
            df = pd.DataFrame(self.blind_spot_data["Average"], index=[0]).T
        if blind_spot_args:
            available_metrics = df.index.tolist()
            valid_args = [arg for arg in blind_spot_args if arg in available_metrics]

            if len(valid_args) < len(blind_spot_args):
                missing_args = set(blind_spot_args) - set(valid_args)
            df = df.loc[valid_args]

        plt.figure(figsize=(14, 8))

        ax = df.T.plot(kind="bar", width=0.75, edgecolor="black", colormap="coolwarm")

        plt.title(
            "Blind Spot Metrics Comparison", fontsize=20, fontweight="bold", pad=20
        )
        plt.xlabel("Metrics", fontsize=14, labelpad=15)
        plt.ylabel("Values", fontsize=14, labelpad=15)
        plt.xticks(rotation=45, ha="right", fontsize=12)
        plt.grid(True, axis="y", linestyle="--", alpha=0.7)
        plt.legend(
            title="Class",
            loc="upper left",
            bbox_to_anchor=(1.05, 1),
            fontsize=12,
            title_fontsize=14,
        )

        for container in ax.containers:
            ax.bar_label(container, fmt="%.2f", padding=3, fontsize=12)
        plt.tight_layout()

        if save_path:
            plt.savefig(save_path, bbox_inches="tight", dpi=300)
        else:
            plt.savefig("blind_spot_metrics_plot.png", bbox_inches="tight", dpi=300)

        plt.show()
        plt.close()

    def _plot_by_meta_data(self, meta_data_args=None, save_path=None):
        """
        Create a bar plot visualization for metrics grouped by metadata.

        :param meta_data_args: (list, optional) List of specific metrics to plot. If None, all metrics will be plotted.
        :param save_path: (str, optional) Path where the plot should be saved. If None, saves as 'metrics_by_meta_data.png'.

        :return: None
        :raises AttributeError: If no valid metadata plot data is loaded.
        """
        if not hasattr(self, "plot_by_meta_data"):
            print("No valid 'plot_by_meta_data' data found in the JSON.")
            return

        data = self.plot_by_meta_data.get("data", {})
        df = pd.DataFrame(data).T

        if meta_data_args:
            df = df[meta_data_args]

        fig, ax = plt.subplots(figsize=(14, 10))

        df.plot(
            kind="bar",
            ax=ax,
            width=0.8,
            edgecolor="black",
            color=sns.color_palette("rocket", n_colors=len(df.columns)),
        )
        ax.set_title(
            f'Metrics for {", ".join(meta_data_args)}',
            fontsize=18,
            fontweight="bold",
            pad=20,
        )
        ax.set_xlabel("Metrics", fontsize=14)
        ax.set_ylabel("Values", fontsize=14)
        ax.tick_params(axis="x", rotation=45)
        ax.grid(True, axis="y", linestyle="--", alpha=0.7)

        for container in ax.containers:
            ax.bar_label(container, fmt="%.2f", padding=3, fontsize=12)

        plt.tight_layout()

        if save_path:
            plt.savefig(save_path, bbox_inches="tight", dpi=300)
        else:
            plt.savefig("metrics_by_meta_data.png", bbox_inches="tight", dpi=300)

        plt.show()
        plt.close()



import pandas as pd
import numpy as np
import pickle
import math
import os

from ..metrics import COCOMetrics


class PlotTopMisses:
    def __init__(
        self,
        ground_truth_dict,
        prediction_dict,
        class_mappings,
        artifacts_path,
        study_list=None,
    ):
        self.ground_truth_dict = ground_truth_dict
        self.prediction_dict = prediction_dict
        self.class_mappings = class_mappings
        self.artifacts_path = artifacts_path
        self.study_list = study_list

    def top_misses(self, metric, sort_by, top_k=150):
        try:
            artifacts = self._load_pickle(self.artifacts_path)
        except:
            coco_metrics = COCOMetrics(self.class_mappings)
            artifacts = coco_metrics.create_artifacts(
                self.ground_truth_dict, self.prediction_dict
            )

        additional_metrics = artifacts.keys()
        top_misses_data = []
        i = 1  # FIX.

        if metric in additional_metrics:
            for image_id, value in (
                pd.DataFrame(artifacts["{}".format(metric)])
                .sort_values(
                    "imagewise_{}".format(metric), ascending=(sort_by == "Ascending")
                )
                .head(top_k)
                .to_dict()["imagewise_{}".format(metric)]
                .items()
            ):

                top_loss_single_image = {
                    "image_id": image_id,
                    "rank": i,
                    metric: value,
                }
                top_misses_data.append(top_loss_single_image)
                i += 1
        else:
            for image_id, value in (
                pd.DataFrame(artifacts["misevals"]["imagewise_metrics"])
                .T.sort_values(metric, ascending=(sort_by == "Ascending"))
                .head(top_k)[metric]
                .to_dict()
                .items()
            ):
                top_loss_single_image = {
                    "image_id": image_id,
                    "rank": i,
                    metric: value,
                }
                top_misses_data.append(top_loss_single_image)
                i += 1

        payload_dict = {"type": "image", "data": top_misses_data}
        return payload_dict

    def _load_pickle(self, file_path):
        with open(file_path, "rb") as f:
            return pickle.load(f)



import math

import numpy as np
import pandas as pd
import sklearn
from sklearn.metrics import auc
from sklearn.preprocessing import label_binarize
from sklearn.metrics import precision_recall_curve, roc_curve, average_precision_score

from ..metrics.auc import AUC
from gesund.core._utils import ValidationUtils, Statistics


class PlotAUC:
    def __init__(self, pred_logits, meta_pred_true, class_mappings):
        self.pred_logits = pred_logits
        self.class_mappings = class_mappings
        self.meta_pred_true = meta_pred_true

        self.validation_utils = ValidationUtils(meta_pred_true)
        self.auc = AUC(class_mappings=class_mappings)

    def precision_recall_multiclass_statistics(self, target_attribute_dict=None):
        """
        Plots ROC Curve for target_class.
        References:
        https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html
        https://plotly.com/python/roc-and-pr-curves/
        :param target_class: target class to produce ROC plot
        :return: payload_dict
        """
        filtered_meta_pred_true = self.validation_utils.filter_attribute_by_dict(
            target_attribute_dict
        )
        true = filtered_meta_pred_true["true"]
        filtered_pred_logits = self.pred_logits.T.loc[
            filtered_meta_pred_true["true"].index
        ].T
        (points, aucs,) = self.auc.calculate_multiclass_precision_recall_statistics(
            true, filtered_pred_logits
        )

        payload_dict = {"type": "precision", "data": {"points": points, "aucs": aucs}}
        return payload_dict

    def roc_multiclass_statistics(self, target_attribute_dict=None):
        """
        Plots ROC Curve for target_class, returns either plotly or matplotlib object.
        References:
        https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html
        https://plotly.com/python/roc-and-pr-curves/
        :param target_class: target class to produce ROC plot
        :param true: True labels.
        :param pred_logits: prediction logits.
        :return: matplotlib or plotly fig object
        """
        filtered_meta_pred_true = self.validation_utils.filter_attribute_by_dict(
            target_attribute_dict
        )
        true = filtered_meta_pred_true["true"]
        filtered_pred_logits = self.pred_logits.T.loc[
            filtered_meta_pred_true["true"].index
        ].T

        points, aucs = self.auc.calculate_multiclass_roc_statistics(
            true=true, pred_logits=filtered_pred_logits, return_points=True
        )

        payload_dict = {"type": "roc", "data": {"points": points, "aucs": aucs}}
        return payload_dict

    def confidence_histogram_scatter_distribution(
        self, predicted_class="overall", n_samples=300, randomize_x=True, n_bins=25
    ):
        # Plot Scatters

        # Filtering data
        pred_true = self.meta_pred_true[["pred_categorical", "true"]].copy()

        if predicted_class != "overall":
            pred_true = pred_true[pred_true["pred_categorical"] == predicted_class]

        if n_samples > pred_true.shape[0]:
            n_samples = pred_true.shape[0]
        pred_true = pred_true.sample(n_samples, replace=True)

        # Filter and obtain probabilities
        pred_true["labels"] = pred_true["pred_categorical"] == pred_true["true"]
        pred_true["labels"] = (
            pred_true["labels"].replace(True, "TP").replace(False, "FP")
        )

        filtered_pred_logits = self.pred_logits[pred_true.index].max()
        pred_true["y"] = filtered_pred_logits

        # Renaming columns
        int_class_mappings = {int(k): v for k, v in self.class_mappings.items()}
        pred_true = pred_true.replace(
            {"pred_categorical": int_class_mappings, "true": int_class_mappings}
        )
        pred_true = pred_true.rename(
            columns={"pred_categorical": "Prediction", "true": "Ground Truth"}
        )

        # Randomize x if needed
        if randomize_x:
            pred_true["x"] = np.random.uniform(0, 1, pred_true.shape[0])
        else:
            pred_true["x"] = pred_true["y"]

        points = list(
            pred_true.reset_index()
            .rename(columns={"index": "image_id"})
            .T.to_dict()
            .values()
        )

        # Plot histogram

        histogram = Statistics.calculate_histogram(
            pred_true["y"], min_=0, max_=1, n_bins=n_bins
        )

        payload_dict = {
            "type": "mixed",
            "data": {"points": points, "histogram": histogram},
        }

        return payload_dict



import os
import pandas as pd
import warnings
from tqdm import tqdm

from .coco_metrics import PlotCOCOMetrics
from .top_misses import PlotTopMisses
from .iou_distribution import PlotIoUDistribution
from .violin_plot import PlotViolin
from .dataset_stats import PlotDatasetStats


class SemanticSegmentationPlotDriver:
    def __init__(
        self,
        class_mappings,
        ground_truth_dict,
        prediction_dict,
        batch_job_id,
        meta_data_dict=None,
        loss_dict=None,
        study_list=None,
        filtering_meta=None,
    ):

        # Create validation variables
        self.true = ground_truth_dict
        self.pred = prediction_dict
        self.class_mappings = class_mappings
        self.meta = None
        self.batch_job_id = batch_job_id
        self.filtering_meta = filtering_meta
        if meta_data_dict:
            self.meta = pd.DataFrame(meta_data_dict).T

        self.sample_size = len(self.true)
        self.class_order = list(range(len(class_mappings.keys())))

        self.artifacts_path = None

        # Import Classes
        self.plot_coco_metrics = PlotCOCOMetrics(
            ground_truth_dict=self.true,
            meta=self.meta,
            prediction_dict=self.pred,
            class_mappings=class_mappings,
            artifacts_path=self.artifacts_path,
            study_list=study_list,
        )

        self.plot_top_misses_ = PlotTopMisses(
            ground_truth_dict=self.true,
            prediction_dict=self.pred,
            class_mappings=self.class_mappings,
            artifacts_path=self.artifacts_path,
            study_list=study_list,
        )

        self.plot_iou_distribution_ = PlotIoUDistribution(
            ground_truth_dict=self.true,
            prediction_dict=self.pred,
            class_mappings=self.class_mappings,
            artifacts_path=self.artifacts_path,
            study_list=study_list,
        )

        self.plot_violin_ = PlotViolin(
            ground_truth_dict=self.true,
            prediction_dict=self.pred,
            class_mappings=self.class_mappings,
            artifacts_path=self.artifacts_path,
            study_list=study_list,
        )

        self.plot_dataset_stats = PlotDatasetStats(meta=self.meta)

    def plot_highlighted_overall_metrics(self):
        return self.plot_coco_metrics.highlighted_overall_metrics()

    def plot_main_metric(self):
        return self.plot_coco_metrics.main_metric()

    def plot_statistics_classbased_table(self, target_attribute_dict=None):
        return self.plot_coco_metrics.statistics_classbased_table(
            target_attribute_dict=self.filtering_meta
        )

    def plot_top_misses(self, metric="IoU", sort_by="Ascending", top_k=150):
        return self.plot_top_misses_.top_misses(
            metric=metric, sort_by=sort_by, top_k=top_k
        )

    def plot_metrics_by_meta_data(self, target_attribute_dict=None):
        return self.plot_coco_metrics.metrics_by_meta_data(
            target_attribute_dict=self.filtering_meta
        )

    def plot_iou_distribution(self, n_samples=300):
        return self.plot_iou_distribution_.iou_distribution(n_samples=n_samples)

    def plot_meta_distribution(self):
        return self.plot_dataset_stats.meta_distributions()

    def plot_violin_graph(self):
        return self.plot_violin_.violin_graph()

    def plot_blind_spot_metrics(self, target_attribute_dict=None):
        return self.plot_coco_metrics.blind_spot_metrics(
            target_attribute_dict=self.filtering_meta, threshold=0.1
        )

    def _calling_all_plots(self):
        # Getting all methods that do not start with '_'
        methods = [
            method_name
            for method_name in dir(self)
            if callable(getattr(self, method_name)) and not method_name.startswith("_")
        ]
        results = {}

        for method_name in tqdm(methods, desc="Calling all plot functions"):
            method = getattr(self, method_name)
            try:
                # Suppress warnings
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")

                    print(f"Calling method: {method_name}...")
                    # Attempt to call the method, handle cases where no arguments are required
                    result = method()
                    results[method_name] = result
            except TypeError as e:
                results[method_name] = f"Could not call {method_name}: {str(e)}"
            except Exception as e:
                results[method_name] = f"Error in {method_name}: {str(e)}"

        return results



import numpy as np
import pandas as pd
from sklearn.metrics import auc
import sklearn

from gesund.core._utils import ValidationUtils, Statistics


class PlotDatasetStats:
    def __init__(self, meta):
        self.meta = meta

    def meta_distributions(self):
        meta_counts = self._calculate_meta_distributions(self.meta)
        data_dict = {}
        data_dict["Validation"] = meta_counts
        payload_dict = {}
        payload_dict["type"] = "bar"
        payload_dict["data"] = data_dict
        return payload_dict

    def _calculate_meta_distributions(self, meta):
        """
        Calculates statistics on meta data.
        :param true: true labels as a list = [1,0,3,4] for 4 sample dataset
        :param pred_categorical: categorical predictions as a list = [1,0,3,4] for 4 sample dataset
        :param labels: order of classes inside list
        :return: dict that contains class dist. for validation/train dataset.
        """
        # Histogram charts for numerical values
        numerical_columns = [
            column
            for column in meta.columns
            if ValidationUtils.is_list_numeric(meta[column].values.tolist())
        ]

        histograms = {
            numerical_column: Statistics.calculate_histogram(
                meta[numerical_column],
                min_=meta[numerical_column].min(),
                max_=meta[numerical_column].max(),
                n_bins=10,
            )
            for numerical_column in numerical_columns
        }

        # Bar charts for categorical values

        categorical_columns = list(set(meta.columns) - set(numerical_columns))
        bars = {
            categorical_column: meta[categorical_column].value_counts().to_dict()
            for categorical_column in categorical_columns
        }

        return {"bar": bars, "histogram": histograms}



import numpy as np
import pandas as pd
from sklearn.metrics import auc
import sklearn
from typing import Any, Dict, List, Optional, Union


class TopLosses:
    def __init__(self, loss: pd.DataFrame, meta_pred_true: pd.DataFrame) -> None:
        self.loss = loss
        self.meta_pred_true = meta_pred_true

    def calculate_top_losses(
        self, predicted_class: Optional[int] = None, top_k: int = 10
    ) -> pd.DataFrame:
        if predicted_class:
            pred_categorical_target_class_index = self.meta_pred_true[
                self.meta_pred_true["pred_categorical"] == predicted_class
            ].index
            sorted_top_loss = self.loss[
                pred_categorical_target_class_index
            ].T.sort_values(by=0, ascending=False)
        else:
            sorted_top_loss = self.loss.T.sort_values(by=0, ascending=False)
        return sorted_top_loss



import argparse
import json
import os
import bson
from pathlib import Path


def read_json(file_path):
    """
    Read a JSON file and return its contents as a Python dictionary.

    This function opens the specified file, reads the JSON content, and
    returns it as a Python dictionary. The file should contain valid JSON data.

    :param file_path: (str) The path to the JSON file to read.

    :return: (dict) The JSON data loaded into a Python dictionary.

    :raises FileNotFoundError: If the specified file does not exist.
    :raises json.JSONDecodeError: If the file contains invalid JSON data.
    """

    with open(file_path, "r") as file:
        data = json.load(file)
    return data


def save_plot_metrics_as_json(overall_metrics, output_dir):
    """
    Save each plot's metrics as individual JSON files in the specified directory.

    This function iterates over the overall metrics dictionary and saves
    each plot's metrics in separate JSON files, named according to the plot names.
    If the directory does not exist, it will be created.

    :param overall_metrics: (dict) A dictionary containing the metrics for multiple plots.
    :param output_dir: (str) Path to the directory where the JSON files should be saved.

    :return: None

    :raises OSError: If the directory cannot be created or if the files cannot be written.
    """

    # Create output directory if it doesn't exist
    Path(output_dir).mkdir(parents=True, exist_ok=True)

    # Iterate over the overall metrics and save each as a separate JSON
    for plot_name, metrics in overall_metrics.items():
        output_file = os.path.join(output_dir, f"{plot_name}.json")
        try:
            with open(output_file, "w") as f:
                json.dump(metrics, f, indent=4)
        except Exception as e:
            print(f"Could not save metrics for {plot_name} because: {e}")


import numpy as np
import pandas as pd


class ValidationUtils:
    def __init__(self, meta_pred_true):
        """
        Initialize the ValidationUtils class.

        :param meta_pred_true: (dict or pd.DataFrame) A dictionary or DataFrame containing metadata and predictions.

        :return: None
        """
        if isinstance(meta_pred_true, dict):
            self.meta_pred_true = pd.DataFrame(meta_pred_true).T
        else:
            self.meta_pred_true = meta_pred_true

    def filter_attribute_by_dict(self, target_attribute_dict=None):
        """
        Filters data by more than one attribute.

        :param target_attribute_dict: (dict, optional) A dictionary where keys are attribute names
                                      and values are the desired filter criteria.

        :return: (pd.DataFrame) A DataFrame containing the filtered metadata based on the specified attributes.
        """
        if bool(target_attribute_dict):
            all_params = target_attribute_dict.keys()
            filtered_meta_pred_true = self.meta_pred_true.copy()
            for target_attribute in all_params:
                if self.is_list_numeric(self.meta_pred_true[target_attribute].values):
                    slider_min, slider_max = target_attribute_dict[target_attribute]
                    filtered_meta_pred_true = filtered_meta_pred_true[
                        self.meta_pred_true[target_attribute].between(
                            slider_min, slider_max
                        )
                    ]
                else:
                    target_value = target_attribute_dict[target_attribute]
                    filtered_meta_pred_true = filtered_meta_pred_true[
                        self.meta_pred_true[target_attribute] == target_value
                    ]
            return filtered_meta_pred_true
        else:
            return self.meta_pred_true

    @staticmethod
    def filter_validation_collection_by_meta(validation_collection_data, meta_filter):
        """
        Filters validation collection data based on specified metadata.

        :param validation_collection_data: (list) A list of validation data documents.
        :param meta_filter: (dict) A dictionary of metadata filters.

        :return: (list) A list of filtered validation collection data documents.
        """
        filtered_validation_collection_data = []
        allowed_meta_keys = list(meta_filter.keys())
        for doc in validation_collection_data:
            add = ValidationUtils.is_meta_in_range(
                meta_data=doc.meta_data, meta_filter=meta_filter
            )
            if add:
                for meta_name in list(
                    set(doc.meta_data.keys()) - set(allowed_meta_keys)
                ):
                    doc.meta_data.pop(meta_name)
                filtered_validation_collection_data.append(doc.dict())
        return filtered_validation_collection_data

    @staticmethod
    def is_meta_in_range(meta_data, meta_filter):
        """
        Determines if the metadata values fall within the specified filter ranges.

        :param meta_data: (dict) A dictionary containing metadata.
        :param meta_filter: (dict) A dictionary containing filter criteria.

        :return: (bool) True if metadata values are within specified ranges, False otherwise.
        """
        add = True
        for meta_key in meta_filter:
            if isinstance(meta_filter[meta_key], str):
                if meta_data[meta_key] != meta_filter[meta_key]:
                    add = False
            elif isinstance(meta_filter[meta_key], list):
                if meta_filter[meta_key][0] == "str":
                    add_multi = []
                    if isinstance(meta_filter[meta_key], str):
                        if meta_data[meta_key] != meta_filter[meta_key]:
                            add_multi.append(False)
                    add = any(add_multi)
                elif not (
                    min(meta_filter[meta_key])
                    <= meta_data[meta_key]
                    <= max(meta_filter[meta_key])
                ):
                    add = False
        return add

    def filter_attribute(self, target_attribute_dict):
        """
        Filters data by a single attribute.

        :param target_attribute_dict: (dict) A dictionary where the key is the attribute name
                                       and the value is the desired filter criteria.

        :return: (pd.DataFrame) A DataFrame containing the filtered metadata based on the specified attribute.
        """
        target_attribute = list(target_attribute_dict.keys())[0]
        target_value = target_attribute_dict[target_attribute]
        if self.is_list_numeric(self.meta_pred_true[target_attribute].values):
            slider_min, slider_max = target_attribute_dict[target_attribute]
            filtered_meta_pred_true = self.meta_pred_true[
                self.meta_pred_true[target_attribute].between(slider_min, slider_max)
            ]
        else:
            filtered_meta_pred_true = self.meta_pred_true[
                self.meta_pred_true[target_attribute] == target_value
            ]
        return filtered_meta_pred_true

    def multifilter_attribute(self, target_attributes_dict):
        """
        Filters data by multiple attributes.

        :param target_attributes_dict: (dict) A dictionary where keys are attribute names
                                        and values are the desired filter criteria.

        :return: (pd.DataFrame) A DataFrame containing the filtered metadata based on the specified attributes.
        """
        all_params = target_attributes_dict.keys()
        filtered_meta_pred_true = self.meta_pred_true.copy()
        for target_attribute in all_params:
            if self.is_list_numeric(self.meta_pred_true[target_attribute].values):
                slider_min, slider_max = target_attributes_dict[target_attribute]
                filtered_meta_pred_true = filtered_meta_pred_true[
                    self.meta_pred_true[target_attribute].between(
                        slider_min, slider_max
                    )
                ]
            else:
                target_value = target_attributes_dict[target_attribute]
                filtered_meta_pred_true = filtered_meta_pred_true[
                    self.meta_pred_true[target_attribute] == target_value
                ]
        return filtered_meta_pred_true

    @staticmethod
    def is_list_numeric(x_list):
        """
        Checks if all elements in the list are numeric.

        :param x_list: (list) A list of elements to check.

        :return: (bool) True if all elements are numeric, False otherwise.
        """
        return all(
            [
                isinstance(
                    i,
                    (
                        int,
                        float,
                        np.int,
                        np.int8,
                        np.int16,
                        np.int32,
                        np.float,
                        np.float16,
                        np.float32,
                        np.float64,
                    ),
                )
                for i in x_list
            ]
        )

    @staticmethod
    def polygon_to_mask(poly, shape, max_value=1):
        """
        Converts a polygon defined by a list of points into a binary mask.

        :param poly: (list) A list of points defining the polygon.
        :param shape: (tuple) The shape of the output mask (height, width).
        :param max_value: (int, optional) The value to fill inside the polygon. Default is 1.

        :return: (np.ndarray) A binary mask with the polygon filled in.
        """
        import skimage

        if isinstance(shape, int):
            img = np.zeros((shape, shape, 1), "uint8")
        else:
            shape = int(shape[0]), int(shape[1]), 1
            img = np.zeros(shape)
        xs = [xy["x"] for xy in poly]
        ys = [xy["y"] for xy in poly]
        # fill polygon

        rr, cc = skimage.draw.polygon(xs, ys, img.shape)
        img[rr, cc] = max_value
        return img

    @staticmethod
    def rle_to_mask(mask_rle: str, shape, label=1):
        """
        Converts run-length encoded mask to a binary mask.

        :param mask_rle: (str) Run-length encoded mask as a string (start length).
        :param shape: (tuple) The shape (height, width) of the output mask.
        :param label: (int, optional) The label to assign to the mask pixels. Default is 1.

        :return: (np.ndarray) A binary mask as a numpy array (1 - mask, 0 - background).
        """
        s = mask_rle.split()
        starts, lengths = [np.asarray(x, dtype=int) for x in (s[0:][::2], s[1:][::2])]
        starts -= 1
        ends = starts + lengths
        img = np.zeros(shape[0] * shape[1], dtype=np.uint8)
        for lo, hi in zip(starts, ends):
            img[lo:hi] = label
        return img.reshape(shape)  # Needed to align to RLE direction

    @staticmethod
    def mask_to_rle(image):
        """
        Converts a binary mask to run-length encoded format.

        :param image: (np.ndarray) A binary mask as a numpy array (1 - mask, 0 - background).

        :return: (tuple) A tuple containing the run-length encoded string and the image shape.
        """
        image_shape = image.shape
        pixels = image.flatten()
        pixels = np.concatenate([[0], pixels, [0]])
        runs = np.where(pixels[1:] != pixels[:-1])[0] + 1
        runs[1::2] -= runs[::2]
        rle_string = " ".join(str(x) for x in runs)
        return rle_string, image_shape

    @staticmethod
    def polygon_to_rle(polygon: list, shape):
        import cv2
        from pycocotools.mask import encode

        """
        Convert a polygon to a Run-Length Encoding (RLE) format mask.

        This function fills a polygon in a mask of the given shape and encodes
        it into Run-Length Encoding format for efficient storage and transmission.

        :param polygon: (list) A list of points representing the vertices of the polygon in the format
                        [[x1, y1], [x2, y2], ...].
        :param shape: (tuple) The shape of the mask to be created, typically in the form
                      (height, width).

        :return: (dict) A dictionary representing the Run-Length Encoding of the mask.

        :raises ValueError: If the polygon points do not form a valid shape or if the shape is invalid.
        """
        mask = np.zeros(shape, dtype=np.uint8)
        polygon = np.asarray(polygon)
        polygon = polygon.reshape(-1, 2)
        cv2.fillPoly(mask, [polygon], 1)

        rle = encode(np.asfortranarray(mask))  # Encoding the mask into RLE
        rle["counts"] = rle["counts"].decode("utf-8")
        return rle

    @staticmethod
    def calculate_iou(gt_mask, pred_mask, threshold=0.5):
        """
        Calculate the Intersection over Union (IoU) between two masks.

        This function computes the IoU score, which is a measure of overlap
        between two masks. The IoU is calculated as the area of intersection
        divided by the area of union.

        :param gt_mask: (list) A list of points representing the ground truth mask vertices
                         in the format [[x1, y1], [x2, y2], ...].
        :param pred_mask: (list) A list of points representing the predicted mask vertices
                          in the format [[x1, y1], [x2, y2], ...].
        :param threshold: (float, optional) The threshold value for IoU, default is 0.5.

        :return: (float) The IoU score in the range [0, 1]. A score of 0 indicates no overlap,
                 while a score of 1 indicates perfect overlap.

        :raises ValueError: If the input masks are invalid or not in the expected format.
        """
        # Don't even convert polygons to mask if there's no intersection
        gt_xs, gt_ys = [i["x"] for i in gt_mask], [i["y"] for i in gt_mask]
        min_gt_x, max_gt_x = min(gt_xs), max(gt_xs)
        min_gt_y, max_gt_y = min(gt_ys), max(gt_ys)

        pred_xs, pred_ys = [i["x"] for i in pred_mask], [i["y"] for i in pred_mask]
        min_pred_x, max_pred_x = min(pred_xs), max(pred_xs)
        min_pred_y, max_pred_y = min(pred_ys), max(pred_ys)

        shape = max(max_pred_x, max_gt_x, max_gt_y, max_pred_y)

        is_intersect = (max(min_gt_x, min_pred_x) < min(max_gt_x, max_pred_x)) and (
            max(min_gt_y, min_pred_y) < min(max_gt_y, max_pred_y)
        )
        if not is_intersect:
            iou = 0
            return iou

        mask1 = ValidationUtils.polygon_to_mask(gt_mask, shape=shape)
        mask2 = ValidationUtils.polygon_to_mask(pred_mask, shape=shape)

        mask1_area = np.count_nonzero(mask1 == 1)
        mask2_area = np.count_nonzero(mask2 == 1)
        intersection = np.count_nonzero(np.logical_and(mask1 == 1, mask2 == 1))
        iou = intersection / (mask1_area + mask2_area - intersection)
        return iou


class Statistics:
    @staticmethod
    def calculate_confidence_interval(metric, len_, z_value=1.96):
        """
        Calculate the confidence interval for a given metric.

        This function computes the confidence interval for a given metric based
        on the provided sample size and z-score. The confidence interval is
        calculated using the formula: CI = metric ± z * sqrt((metric * (1 - metric)) / len_).

        :param metric: (float) The observed metric for which to calculate the confidence interval.
        :param len_: (int) The sample size used to estimate the confidence interval.
        :param z_value: (float, optional) The z-score corresponding to the desired confidence level,
                        default is 1.96 (for 95% confidence level).

        :return: (tuple) A tuple containing the lower and upper bounds of the confidence interval.

        :raises ValueError: If the sample size is less than or equal to zero.
        """
        metric = np.abs(metric)
        ci_length = z_value * np.sqrt((metric * (1 - metric)) / len_)
        ci_lower = metric - ci_length
        ci_upper = metric + ci_length
        return (ci_lower, ci_upper)

    @staticmethod
    def calculate_histogram(array_, min_, max_, n_bins):
        """
        Calculate a histogram for a given array.

        This function computes a histogram for the specified array, creating bins
        based on the provided minimum and maximum values. The histogram is returned
        as a list of dictionaries with bin categories and counts.

        :param array_: (list or np.ndarray) The array of values for which to calculate the histogram.
        :param min_: (float) The minimum value for the histogram bins.
        :param max_: (float) The maximum value for the histogram bins.
        :param n_bins: (int) The number of bins to create in the histogram.

        :return: (list) A list of dictionaries where each dictionary contains a category
                 and the corresponding value count for that bin.

        :raises ValueError: If n_bins is less than or equal to zero.
        """
        array = np.array(array_)
        bin_spaces = np.linspace(min_, max_, n_bins + 1)
        histogram_list = list()

        for i in range(len(bin_spaces) - 1):
            bin_min = bin_spaces[i]
            bin_max = bin_spaces[i + 1]
            histogram_list.append(
                {
                    "category": f"{bin_min.round(2)}",
                    "value": np.sum([(bin_min < array_) & (array_ <= bin_max)]).item(),
                }
            )

        return histogram_list



import math

import numpy as np
import pandas as pd
import sklearn
from sklearn.metrics import auc
from sklearn.preprocessing import label_binarize
from sklearn.metrics import precision_recall_curve, roc_curve, average_precision_score

from ..metrics.auc import AUC
from ..metrics.stats_tables import *
from ..plots.stats_tables import PlotStatsTables


class PlotBlindSpot:
    def __init__(
        self, true, pred_logits, pred_categorical, class_mappings, meta_pred_true
    ):
        self.true = true
        self.pred_logits = pred_logits
        self.pred_categorical = pred_categorical
        self.class_mappings = class_mappings
        self.class_order = [int(i) for i in list(class_mappings.keys())]

        self.plot_stats_tables = PlotStatsTables(
            true=true,
            pred_logits=pred_logits,
            pred_categorical=pred_categorical,
            meta_pred_true=meta_pred_true,
            class_mappings=class_mappings,
        )

        self.auc = AUC(class_mappings)

    def blind_spot_metrics(self, target_attribute_dict=None):
        """
        Plots ROC Curve for target_class.
        References:
        https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html
        https://plotly.com/python/roc-and-pr-curves/
        :param target_class: target class to produce ROC plot
        :return: payload_dict
        """
        blind_spot_metrics_dict = self.plot_stats_tables.blind_spot_metrics(
            target_attribute_dict
        )

        return blind_spot_metrics_dict



import pickle

import numpy as np
import pandas as pd

from ..metrics.coco_metrics import COCOMetrics
from gesund.core._utils import ValidationUtils, Statistics


class PlotCOCOMetrics:
    def __init__(
        self,
        class_mappings,
        meta,
        ground_truth_dict=None,
        prediction_dict=None,
        artifacts_path=None,
        study_list=None,
    ):
        if ground_truth_dict:
            self.ground_truth_dict = ground_truth_dict
            self.prediction_dict = prediction_dict
            self.coco_metrics = COCOMetrics(class_mappings, study_list)
            self.study_list = study_list
            self.class_mappings = class_mappings
            self.validation_utils = ValidationUtils(meta)
            self.artifacts_path = artifacts_path

    def highlighted_overall_metrics(self):
        metrics_dict = self.coco_metrics.calculate_highlighted_overall_metrics(
            gt=self.ground_truth_dict, pred=self.prediction_dict
        )
        metrics_dict = {k: {"Validation": v} for k, v in metrics_dict.items()}
        payload_dict = {"type": "overall", "data": metrics_dict}
        return payload_dict

    def study_classbased_table(self, target_attribute_dict):

        if target_attribute_dict:
            gt, pred = self.validation_utils._filter_dict_by_study(
                self.ground_truth_dict,
                self.prediction_dict,
                target_attribute_dict["study_id"],
            )
            metrics_dict = self.coco_metrics.calculate_statistics_classbased_table(
                gt=gt, pred=pred, target_attribute_dict=target_attribute_dict
            )

        else:
            metrics_dict = self.coco_metrics.calculate_statistics_classbased_table(
                gt=self.ground_truth_dict,
                pred=self.prediction_dict,
                target_attribute_dict=target_attribute_dict,
            )

        payload_dict = {"type": "table", "data": {"Validation": metrics_dict}}

        return payload_dict

    def statistics_classbased_table(self, target_attribute_dict):

        if self.study_list:
            return {}

        if target_attribute_dict:

            filtered_image_rows = self.validation_utils.filter_attribute_by_dict(
                target_attribute_dict
            )
            filtered_image_ids = filtered_image_rows.index.tolist()

            gt = dict(
                (image_id, self.ground_truth_dict[image_id])
                for image_id in filtered_image_ids
                if image_id in self.ground_truth_dict
            )
            pred = dict(
                (image_id, self.prediction_dict[image_id])
                for image_id in filtered_image_ids
                if image_id in self.prediction_dict
            )

            metrics_dict = self.coco_metrics.calculate_statistics_classbased_table(
                gt=gt, pred=pred, target_attribute_dict=target_attribute_dict
            )

        else:
            metrics_dict = self.coco_metrics.calculate_statistics_classbased_table(
                gt=self.ground_truth_dict,
                pred=self.prediction_dict,
                target_attribute_dict=target_attribute_dict,
            )

        payload_dict = {"type": "table", "data": {"Validation": metrics_dict}}
        return payload_dict

    def metrics_by_meta_data(self, target_attribute_dict=None):
        # Filter by meta
        image_ids = self.validation_utils.filter_attribute_by_dict(
            target_attribute_dict
        ).index.tolist()

        try:
            artifacts = self._load_pickle(self.artifacts_path)
        except:
            coco_metrics = COCOMetrics(self.class_mappings)
            artifacts = self.coco_metrics.create_artifacts(
                self.ground_truth_dict, self.prediction_dict
            )

        imagewise_iou = artifacts["iou"]["imagewise_iou"]
        imagewise_pixel_accuracy = artifacts["pAccs"]["imagewise_acc"]
        imagewise_fwiou = artifacts["fwiou"]["imagewise_fwiou"]
        imagewise_misevals = artifacts["misevals"]["imagewise_metrics"]

        for image_id in imagewise_iou:
            imagewise_misevals[image_id]["iou"] = imagewise_iou[image_id]
            imagewise_misevals[image_id]["pAccs"] = imagewise_iou[image_id]
            imagewise_misevals[image_id]["fwiou"] = imagewise_iou[image_id]

        imagewise_metrics_df = pd.DataFrame(imagewise_misevals)
        imagewise_metrics_df = imagewise_metrics_df.T.loc[image_ids]
        # Delete repetitives
        imagewise_metrics_df = imagewise_metrics_df.drop(
            columns=[
                "IoU",
                "MCC_Normalized",
                "MCC_Absolute",
                "Precision_Sets",
                "Precision_CM",
                "Specificity_Sets",
                "Specificity_CM",
                "Sensitivity_Sets",
                "Sensitivity_CM",
                "IoU_Sets",
                "IoU_CM",
                "DSC_Enhanced",
                "DSC_Sets",
                "DSC_CM",
                "Accuracy",
                "Accuracy_Sets",
                "Accuracy_CM",
            ]
        )
        # Rename
        imagewise_metrics_df = imagewise_metrics_df.rename(
            columns={
                "iou": "meanIoU",
                "Kapp": "mean Kappa",
                "DSC": "Dice Score",
                "Sensitivity": "mean Sensitivity",
                "Specificity": "mean Specificity",
                "AUC": "mean AUC",
                "fwiou": "mean fwIoU",
                "pAccs": "Pixel Accuracy",
                "MCC": "Matthew's Classwise C C",
            }
        )

        metrics = imagewise_metrics_df.mean().to_dict()

        payload_dict = {"type": "bar", "data": {}}

        graph_1_metrics = [
            "TruePositive",
            "TrueNegative",
            "FalsePositive",
            "FalseNegative",
        ]
        graph_2_metrics = ["AverageHausdorffDistance", "SimpleHausdorffDistance"]

        graph_1_dict = {
            key: int(metrics[key]) for key in graph_1_metrics if key in metrics
        }
        payload_dict["data"]["graph_1"] = graph_1_dict

        graph_2_dict = {key: metrics[key] for key in graph_2_metrics if key in metrics}
        payload_dict["data"]["graph_2"] = graph_2_dict

        graph_3_dict = {
            key: metrics[key]
            for key in metrics
            if key not in graph_1_metrics and key not in graph_2_metrics
        }
        payload_dict["data"][
            "graph_3"
        ] = graph_3_dict  # Remaining metrics to graph_3 (specified by frontend)

        return payload_dict

    def main_metric(self):
        metrics_dict = (
            metrics_dict
        ) = self.coco_metrics.calculate_highlighted_overall_metrics(
            gt=self.ground_truth_dict, pred=self.prediction_dict
        )
        payload_dict = {
            "type": "main_metric",
            "data": {
                "mean IoU": metrics_dict["Vol.mIoU"]
                if self.study_list
                else metrics_dict["mean IoU"],
            },
        }
        return payload_dict

    def blind_spot_metrics(self, target_attribute_dict, threshold):
        """
        Plots ROC Curve for target_class.
        References:
        https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html
        https://plotly.com/python/roc-and-pr-curves/
        :param target_class: target class to produce ROC plot
        :return: payload_dict
        """
        filtered_image_ids = self.validation_utils.filter_attribute_by_dict(
            target_attribute_dict
        ).index.tolist()

        filtered_gt = dict(
            (image_id, self.ground_truth_dict[image_id])
            for image_id in filtered_image_ids
            if image_id in self.ground_truth_dict
        )
        filtered_pred = dict(
            (image_id, self.ground_truth_dict[image_id])
            for image_id in filtered_image_ids
            if image_id in self.prediction_dict
        )

        class_metrics = self.coco_metrics.calculate_statistics_classbased_table(
            gt=filtered_gt, pred=filtered_pred
        )

        overall_metrics = self.coco_metrics.calculate_highlighted_overall_metrics(
            gt=filtered_gt, pred=filtered_pred
        )

        blind_spot_metrics_dict = {
            str(i): value for i, value in enumerate(class_metrics.values(), 0)
        }

        blind_spot_metrics_dict["Average"] = overall_metrics

        return blind_spot_metrics_dict

    def create_artifacts(self, artifacts_path):
        status = False
        try:
            self.coco_metrics.create_artifacts(
                gt=self.ground_truth_dict,
                pred=self.prediction_dict,
                artifacts_path=artifacts_path,
            )
        except:
            print(
                f"Artifacts can't be created to, {artifacts_path}, metrics values will be for recalculated in each request"
            )
        return status

    def _load_pickle(self, file_path):
        with open(file_path, "rb") as f:
            return pickle.load(f)



import random
import pickle

import pandas as pd
import numpy as np

from ..metrics.coco_metrics import COCOMetrics
from gesund.core._utils import ValidationUtils, Statistics


class PlotIoUDistribution:
    def __init__(
        self,
        class_mappings,
        ground_truth_dict=None,
        prediction_dict=None,
        artifacts_path=None,
        study_list=None,
    ):
        if ground_truth_dict:
            self.ground_truth_dict = ground_truth_dict
            self.prediction_dict = prediction_dict
            self.class_mappings = class_mappings
            self.coco_metrics = COCOMetrics(class_mappings)
            self.artifacts_path = artifacts_path
            self.study_list = study_list

    def iou_distribution(self, n_samples=300):
        try:
            artifacts = self._load_pickle(self.artifacts_path)
        except:
            coco_metrics = COCOMetrics(self.class_mappings)
            artifacts = coco_metrics.create_artifacts(
                self.ground_truth_dict, self.prediction_dict
            )

        iou_series = pd.Series(artifacts["iou"]["imagewise_iou"]).sort_values()
        iou_series = iou_series.sample(min(n_samples, iou_series.size)).to_dict()

        points = []

        for image_id in iou_series:
            iou_point = {
                "image_id": image_id,
                "x": random.uniform(0, 1),
                "y": float(iou_series[image_id]),
            }
            points.append(iou_point)

        # Plot Histogram
        histogram = Statistics.calculate_histogram(
            pd.Series(iou_series.values()), min_=0, max_=1, n_bins=10
        )

        payload_dict = payload_dict = {
            "type": "mixed",
            "data": {"points": points, "histogram": histogram},
        }

        return payload_dict

    def _load_pickle(self, file_path):
        with open(file_path, "rb") as f:
            return pickle.load(f)



from .accuracy import Accuracy
from .auc import AUC
from .most_confused import MostConfused
from .dataset_stats import DatasetStats
from .stats_tables import StatsTables
from .confusion_matrix import ConfusionMatrix
from .top_losses import TopLosses
from .threshold import ThresholdMetrics
from .lift_chart import LiftChart



import numpy as np
import sklearn
import pandas as pd
from sklearn.metrics import auc
from sklearn.metrics import roc_auc_score

from gesund.core._metrics.classification.metrics.auc import AUC
from gesund.core._metrics.classification.metrics.stats_tables import StatsTables
from gesund.core._utils import ValidationUtils


class PlotStatsTables:
    def __init__(
        self, true, pred_logits, pred_categorical, class_mappings, meta_pred_true
    ):
        self.true = true
        self.pred_logits = pred_logits
        self.pred_categorical = pred_categorical
        self.class_mappings = class_mappings
        self.class_order = [int(i) for i in list(class_mappings.keys())]

        self.stats_tables = StatsTables(
            class_mappings=class_mappings,
        )
        self.auc = AUC(class_mappings)
        self.validation_utils = ValidationUtils(meta_pred_true)

    def training_validation_comparison_classbased_table(self):
        rename_statistics_dict = {
            "FP": "False Positive",
            "TP": "True Positive",
            "FN": "False Negative",
            "TN": "True Negative",
            "TPR": "Recall",
            "TNR": "Specificity (TNR)",
            "PPV": "Precision",
            "NPV": "Negative Predictive Value",
            "FPR": "False Positive Rate",
            "FNR": "False Negative Rate",
            "F1": "F1",
        }

        data = {}
        content_dict = {}
        (
            accuracy,
            micro_f1,
            macro_f1,
            macro_auc,
            micro_prec,
            macro_prec,
            micro_rec,
            macro_rec,
            macro_specificity,
            micro_specificity,
            matthews,
        ) = self.stats_tables.calculate_highlighted_overall_metrics(
            true=self.true,
            pred_categorical=self.pred_categorical,
            pred_logits=self.pred_logits,
        )

        data = {
            "accuracy": {"Validation": accuracy},
            "micro_f1": {"Validation": micro_f1},
            "macro_f1": {"Validation": macro_f1},
            "macro_auc": {"Validation": macro_auc},
            "micro_precision": {"Validation": micro_prec},
            "macro_precision": {"Validation": macro_prec},
            "micro_recall": {"Validation": micro_rec},
            "macro_recall": {"Validation": macro_rec},
        }
        payload_dict = {"type": "table", "data": data}
        return payload_dict

    def tp_tn_fp_fn(self, target_class=None, target_attribute_dict=None):
        """
        Calculates true positive, true negative, false positive, false negatives for the given class.
        :param target_class: Class to calculate metrics.
        :param true: True labels for samples
        :param pred_categorical: Prediction for samples
        :return: payload_dict
        """
        filtered_meta_pred_true = self.validation_utils.filter_attribute_by_dict(
            target_attribute_dict
        )
        true = filtered_meta_pred_true["true"]
        pred_categorical = filtered_meta_pred_true["pred_categorical"]

        sense_spec_dict = self.auc.calculate_sense_spec(true, pred_categorical)

        TP = sense_spec_dict["TP"]
        TN = sense_spec_dict["TN"]
        FP = sense_spec_dict["FP"]
        FN = sense_spec_dict["FN"]

        if target_class is not None:
            TP = TP[target_class]
            TN = TN[target_class]
            FP = FP[target_class]
            FN = FN[target_class]

        payload_dict = {
            "type": "square",
            "data": {"TP": TP, "TN": TN, "FP": FP, "FN": FN},
        }
        return payload_dict

    def highlighted_overall_metrics(
        self, target_attribute_dict=None, cal_conf_interval=True
    ):
        payload_dict = {}

        image_ids = self.true.index

        if target_attribute_dict:
            filtered_meta_pred_true = self.validation_utils.filter_attribute_by_dict(
                target_attribute_dict
            )

            image_ids = filtered_meta_pred_true.index

        (
            accuracy,
            micro_f1,
            macro_f1,
            macro_auc,
            micro_prec,
            macro_prec,
            micro_rec,
            macro_rec,
            macro_specificity,
            micro_specificity,
            matthews,
        ) = self.stats_tables.calculate_highlighted_overall_metrics(
            true=self.true.loc[image_ids],
            pred_categorical=self.pred_categorical.loc[image_ids],
            pred_logits=self.pred_logits.T.loc[image_ids].T,
            cal_conf_interval=cal_conf_interval,
        )
        if cal_conf_interval:
            data = {
                "Accuracy": {
                    "Validation": accuracy[0],
                    "Confidence_Interval": list(accuracy[1]),
                },
                "Micro F1": {
                    "Validation": micro_f1[0],
                    "Confidence_Interval": list(micro_f1[1]),
                },
                "Macro F1": {
                    "Validation": macro_f1[0],
                    "Confidence_Interval": list(macro_f1[1]),
                },
                "AUC": {
                    "Validation": macro_auc[0],
                    "Confidence_Interval": list(macro_auc[1]),
                },
                "Precision": {
                    "Validation": macro_prec[0],
                    "Confidence_Interval": list(macro_prec[1]),
                },
                "Sensitivity": {
                    "Validation": macro_rec[0],
                    "Confidence_Interval": list(macro_rec[1]),
                },
                "Specificity": {
                    "Validation": macro_specificity[0],
                    "Confidence_Interval": list(macro_specificity[1]),
                },
                "Matthews C C": {
                    "Validation": matthews[0],
                    "Confidence_Interval": list(matthews[1]),
                },
            }
        else:
            data = {
                "Accuracy": {"Validation": accuracy},
                "Micro F1": {"Validation": micro_f1},
                "Macro F1": {"Validation": macro_f1},
                "AUC": {"Validation": macro_auc},
                "Precision": {"Validation": macro_prec},
                "Sensitivity": {"Validation": macro_rec},
                "Specificity": {"Validation": macro_specificity},
                "Matthews C C": {"Validation": matthews},
            }

        payload_dict["type"] = "overall"
        payload_dict["data"] = data
        return payload_dict

    def statistics_classbased_table(self, target_attribute_dict=None):
        """
        Calculates True Positive,True Negative, False Positive, False Negative for an attribute filtered data.
        :param target_class: Class of interest to calculate ROC.
        :param target_attribute_dict: Dictionary of attribute-value pairs.
        If "Age" in meta_data, target_attribute_dict = {"Age:[10,30]} indicates  Age between 10 and 30.
        If "Gender" in meta_data, target_attribute_dict = {"Gender:"male"] indicates  male accuracy.
        :return: payload_dict
        """
        #  TO DO: Reduce nofilter/filtered/multifiltered functions to single one, which is possible.
        filtered_meta_pred_true = self.validation_utils.filter_attribute_by_dict(
            target_attribute_dict
        )
        true = filtered_meta_pred_true["true"]
        pred_categorical = filtered_meta_pred_true["pred_categorical"]
        sense_spec_dict = self.stats_tables.calculate_statistics_classbased_table(
            true, pred_categorical
        )

        # Get ROC AUC
        pred_logits = self.pred_logits[pred_categorical.index]
        class_aucs = self.auc.calculate_multiclass_roc_statistics(
            true, pred_logits, return_points=False, use_class_name=False
        )

        for key in ["Micro", "Macro"]:
            class_aucs.pop(key, None)
        for key in list(class_aucs.keys()):
            class_aucs[int(key)] = class_aucs.pop(key)

        data = dict()
        stats_table_validation = pd.DataFrame(sense_spec_dict)
        stats_table_validation["AUC"] = class_aucs.values()

        stats_table_validation_renamed = stats_table_validation.rename(
            index={int(k): v for k, v in self.class_mappings.items()}
        ).T.to_dict()
        stats_table_validation_dict_per_row = [
            {"Class": keys, **dict(stats_table_validation_renamed[keys])}
            for keys in stats_table_validation_renamed
        ]

        data["Validation"] = stats_table_validation_renamed

        payload_dict = {"type": "table", "data": data}
        return payload_dict

    def class_performances(self):
        data = {}
        metrics = ["Accuracy", "F1", "TPR", "TNR", "PPV"]
        rename_statistics_dict = {
            "FP": "False Positive",
            "TP": "True Positive",
            "FN": "False Negative",
            "TN": "True Negative",
            "TPR": "Sensitivity",
            "TNR": "Specificity",
            "PPV": "Precision",
            "NPV": "Negative Predictive Value",
            "FPR": "False Positive Rate",
            "FNR": "False Negative Rate",
            "F1": "F1",
        }

        sense_spec_dict = self.auc.calculate_sense_spec(
            self.true, self.pred_categorical
        )

        # Filter designed metrics
        sense_spec_dict = {i: sense_spec_dict[i] for i in metrics}

        # Rename Metrics

        common_metrics = set(sense_spec_dict.keys()).intersection(
            rename_statistics_dict.keys()
        )

        for key in common_metrics:
            sense_spec_dict[str(rename_statistics_dict[key])] = sense_spec_dict[key]
            sense_spec_dict.pop(key)

        # Rename classes
        for key_ in sense_spec_dict:
            for class_ in self.class_order:
                sense_spec_dict[key_][
                    self.class_mappings[str(class_)]
                ] = sense_spec_dict[key_][class_]
                sense_spec_dict[key_].pop(class_)

        data["Validation"] = sense_spec_dict

        payload_dict = {}
        payload_dict["type"] = "bar"
        payload_dict["data"] = data
        return payload_dict

    def blind_spot_metrics(self, target_attribute_dict=None):
        payload_dict = {}
        image_ids = self.true.index

        if target_attribute_dict:
            filtered_meta_pred_true = self.validation_utils.filter_attribute_by_dict(
                target_attribute_dict
            )

            image_ids = filtered_meta_pred_true.index

        blind_spot_metrics_dict = self.stats_tables.calculate_blind_spot_metrics(
            true=self.true.loc[image_ids],
            pred_categorical=self.pred_categorical.loc[image_ids],
            pred_logits=self.pred_logits.T.loc[image_ids].T,
        )

        return blind_spot_metrics_dict



from typing import Optional, Union
import bson
import os
import json

from gesund.core._converters import ConverterFactory
from gesund.core._data_loaders import DataLoader
from gesund.core._schema import (
    UserInputParams,
    UserInputData,
    ResultDataClassification,
    ResultDataObjectDetection,
    ResultDataSegmentation,
)
from gesund.core._plot import PlotData
from gesund.core._exceptions import MetricCalculationError


class ValidationProblemTypeFactory:
    @classmethod
    def get_problem_type_factory(cls, problem_type: str):
        """
        Return the validation creation class based on the problem_type.

        :param problem_type: Type of problem (e.g., 'classification', 'object_detection').
        :type problem_type: str

        :return: Validation creation class corresponding to the problem type.
        :rtype: class
        """
        if problem_type == "classification":
            from gesund.core._metrics.classification.create_validation import (
                ValidationCreation,
            )

            return ValidationCreation
        elif problem_type == "semantic_segmentation":
            from gesund.core._metrics.semantic_segmentation.create_validation import (
                ValidationCreation,
            )

            return ValidationCreation
        elif problem_type == "object_detection":
            from gesund.core._metrics.object_detection.create_validation import (
                ValidationCreation,
            )

            return ValidationCreation
        else:
            raise ValueError(f"Unknown problem type: {problem_type}")


class Validation:
    def __init__(
        self,
        annotations_path: str,
        predictions_path: str,
        class_mapping: Union[str, dict],
        problem_type: str,
        data_format: str,
        json_structure_type: str,
        plot_config: dict,
        metadata_path: Optional[str] = None,
        return_dict: Optional[bool] = False,
        display_plots: Optional[bool] = False,
        store_plots: Optional[bool] = False,
        run_validation_only: Optional[bool] = False,
    ):
        """
        Initialization function to handle the validation pipeline

        :param annotations_path: Path to the JSON file containing the annotations data.
        :type annotations_path: str

        :param predictions_path: Path to the JSON file containing the predictions data.
        :type predictions_path: str

        :param class_mappings: Path to the JSON file containing class mappings or a dictionary file.
        :type class_mappings: Union[str, dict]

        :param problem_type: Type of problem (e.g., 'classification', 'object_detection').
        :type problem_type: str

        :param json_structure_type: Data format for the validation (e.g., 'coco', 'yolo', 'gesund').
        :type json_structure_type: str

        :param plot_config: The configuration in dictionary format required for plotting
        :type plot_config: dict

        :param metadata_path: Path to the metadata file (if available).
        :type metadata_path: str
        :optional metadata_path: true

        :param return_dict: If true then the return the result as dict
        :type return_dict: bool
        :optional return_dict: true

        :param display_plots: if true then the plots are displayed
        :type display_plots: bool
        :optional display_plots: true

        :param store_plots: if true then the plots are saved in local as png files
        :type store_plots: bool
        :optional store_plots: true

        :param run_validation_only: by default true the plots are run
        :type run_validation_only: bool
        :optional run_validation_only: true


        :return: None
        """
        # set up user parameters
        params = {
            "annotations_path": annotations_path,
            "predictions_path": predictions_path,
            "metadata_path": metadata_path,
            "problem_type": problem_type,
            "json_structure_type": json_structure_type,
            "return_dict": return_dict,
            "display_plots": display_plots,
            "store_plots": store_plots,
            "data_format": data_format,
            "class_mapping": class_mapping,
            "plot_config": plot_config,
            "run_validation_only": run_validation_only,
        }
        self.user_params = UserInputParams(**params)

        # set up and load all the required data
        self._load_data()

        # set up batch job id
        self.batch_job_id = str(bson.ObjectId())
        self.output_dir = f"outputs/{self.batch_job_id}"

        os.makedirs(self.output_dir, exist_ok=True)
        os.makedirs(os.path.join(self.output_dir, "plots"), exist_ok=True)
        self.problem_type_result_map = {
            "classification": ResultDataClassification,
            "object_detection": ResultDataObjectDetection,
            "semantic_segmentation": ResultDataSegmentation,
        }

    def _load_data(self) -> dict:
        """
        A Function to load the JSON files

        :return: None
        """
        # Load data
        # set up source data for processing
        data_loader = DataLoader(self.user_params.data_format)
        data = {
            "prediction": data_loader.load(self.user_params.predictions_path),
            "annotation": data_loader.load(self.user_params.annotations_path),
        }
        if isinstance(self.user_params.class_mapping, str):
            data["class_mapping"] = data_loader.load(self.user_params.class_mapping)
        else:
            data["class_mapping"] = self.user_params.class_mapping

        if self.user_params.metadata_path:
            data["metadata"] = data_loader.load(self.user_params.metadata_path)

        # run conversion
        if self.user_params.json_structure_type != "gesund":
            self._convert_data(data)

        self.data = UserInputData(**data)

    def _convert_data(self, data):
        """
        A function to convert the data from the respective structure to the gesund format

        :param data: dictionary containing the data
        :type data: dict

        :return: data dictionary
        :rtype: dict
        """
        # setup data converter
        data_converter = ConverterFactory().get_converter(
            self.user_params.json_structure_type
        )

        # run the converters
        (
            data["converted_annotation"],
            data["converted_prediction"],
        ) = data_converter.convert(
            annotation=data["annotation"],
            prediction=data["prediction"],
            problem_type=self.user_params.problem_type,
        )
        data["was_converted"] = True
        return data

    def _run_validation(self) -> dict:
        """
        A function to run the validation

        :param: None

        :return: result dictionary
        :rtype: dict
        """
        _validation_class = ValidationProblemTypeFactory().get_problem_type_factory(
            self.user_params.problem_type
        )

        _metric_validation_executor = _validation_class(self.batch_job_id)
        try:
            if self.data.was_converted:
                prediction = self.data.converted_prediction
                annotation = self.data.converted_annotation
            else:
                prediction = self.data.prediction
                annotation = self.data.annotation
            metadata = None
            if self.user_params.metadata_path:
                metadata = self.data.metadata
            validation_data = (
                _metric_validation_executor.create_validation_collection_data(
                    prediction, annotation, metadata
                )
            )

            metrics = _metric_validation_executor.load(
                validation_data, self.data.class_mapping
            )
            return metrics
        except Exception as e:
            print(e)
            raise MetricCalculationError("Error in calculating metrics!")

    def _save_json(self, results) -> None:
        """
        A function to save the validation results

        :param results: Dictionary containing the results
        :type results: dict

        :return: None
        """
        for plot_name, metrics in results.items():
            output_file = os.path.join(self.output_dir, f"{plot_name}.json")
            try:
                with open(output_file, "w") as f:
                    json.dump(metrics, f, indent=4)
            except Exception as e:
                print(f"Could not save metrics for {plot_name} because: {e}")

    @staticmethod
    def format_metrics(metrics: dict) -> dict:
        """
        Format and print the overall metrics in a readable format.

        This function takes in the metrics data, formats it, and prints out the
        highlighted overall metrics, including confidence intervals when applicable.
        It also prints a message indicating that all graphs and plot metrics have been saved.

        :param metrics: (dict) A dictionary containing the metrics, expected to have
                        a 'plot_highlighted_overall_metrics' key with the metrics data.

        :return: None
        """

        print("\nOverall Highlighted Metrics:\n" + "-" * 40)
        for metric, values in metrics["plot_highlighted_overall_metrics"][
            "data"
        ].items():
            print(f"{metric}:")
            for key, value in values.items():
                if isinstance(value, list):  # If it's a confidence interval
                    value_str = f"{value[0]:.4f} to {value[1]:.4f}"
                else:
                    value_str = f"{value:.4f}"
                print(f"    {key}: {value_str}")
            print("-" * 40)
        print("All Graphs and Plots Metrics saved in JSONs.\n" + "-" * 40)

    def _plot_metrics(self, results: dict) -> None:
        """
        A function to plot the metrics

        :param results: a dictionary containing the validation metrics
        :type results: dict

        :return: None
        """
        plot_data_executor = PlotData(
            metrics_result=results,
            user_params=self.user_params,
            user_data=self.data,
            batch_job_id=self.batch_job_id,
            validation_problem_type_factory=ValidationProblemTypeFactory(),
        )
        plot_data_executor.plot()

    def run(self) -> Union[None, ResultDataClassification]:
        """
        A function to run the validation pipeline

        :param: None
        :type:

        :return: None
        :rtype:
        """
        results = {}

        # run the validation
        results = self._run_validation()

        # format the results
        # results = self._format_results(results)

        # store the results
        self._save_json(results)

        # plot the metrics only if user requires
        if self.user_params.run_validation_only is False:
            self._plot_metrics(results)

        # return the results
        if self.user_params.return_dict:
            results = self.problem_type_result_map[self.user_params.problem_type](
                **results
            )
            return results



import pickle

import pandas as pd
import numpy as np

from ..metrics.coco_metrics import COCOMetrics


class PlotViolin:
    def __init__(
        self,
        class_mappings,
        ground_truth_dict=None,
        prediction_dict=None,
        artifacts_path=None,
        study_list=None,
    ):
        if ground_truth_dict:
            self.ground_truth_dict = ground_truth_dict
            self.prediction_dict = prediction_dict
            self.class_mappings = class_mappings
            self.coco_metrics = COCOMetrics(class_mappings)
            self.artifacts_path = artifacts_path
            self.study_list = study_list

    def violin_graph(self):

        try:
            artifacts = self._load_pickle(self.artifacts_path)
        except:
            coco_metrics = COCOMetrics(self.class_mappings)
            artifacts = coco_metrics.create_artifacts(
                self.ground_truth_dict, self.prediction_dict
            )

        # TO DO: add imagewise APD in its function
        # Add in coco_metrics.py plots as well
        iou_series = pd.DataFrame.from_dict(
            artifacts["iou"]["imagewise_iou"], orient="index", columns=["imagewise_iou"]
        ).sort_values(by="imagewise_iou", ascending=True)
        fwiou_series = pd.DataFrame.from_dict(
            artifacts["fwiou"]["imagewise_fwiou"],
            orient="index",
            columns=["imagewise_fwiou"],
        ).sort_values(by="imagewise_fwiou", ascending=True)
        acc_series = pd.DataFrame.from_dict(
            artifacts["pAccs"]["imagewise_acc"],
            orient="index",
            columns=["imagewise_acc"],
        ).sort_values(by="imagewise_acc", ascending=True)

        dice_series = self.create_miseval_imagewise_df(artifacts, "DSC")
        spec_series = self.create_miseval_imagewise_df(artifacts, "Specificity")
        sens_series = self.create_miseval_imagewise_df(artifacts, "Sensitivity")
        kapp_series = self.create_miseval_imagewise_df(artifacts, "Kapp")
        auc_series = self.create_miseval_imagewise_df(artifacts, "AUC")

        payload_dict = {
            "type": "violin",
            "data": {
                "IoU": iou_series["imagewise_iou"].tolist(),
                "FWIoU": fwiou_series["imagewise_fwiou"].tolist(),
                "Acc": acc_series["imagewise_acc"].tolist(),
                "Dice": dice_series["DSC"].tolist(),
                "Spec": spec_series["Specificity"].tolist(),
                "Sens": sens_series["Sensitivity"].tolist(),
                "Kapp": kapp_series["Kapp"].tolist(),
                "AUC": auc_series["AUC"].tolist(),
            },
        }

        return payload_dict

    def create_miseval_imagewise_df(self, artifacts, metric_name):
        metric_values = [
            (img_id, metrics[metric_name])
            for img_id, metrics in artifacts["misevals"]["imagewise_metrics"].items()
        ]
        df = pd.DataFrame(metric_values, columns=["Image ID", metric_name]).sort_values(
            metric_name
        )
        return df

    def _load_pickle(self, file_path):
        with open(file_path, "rb") as f:
            return pickle.load(f)



import numpy as np
import pandas as pd
import sklearn
from sklearn.metrics import auc
from sklearn.preprocessing import label_binarize
from sklearn.metrics import precision_recall_curve, roc_curve, average_precision_score


class PlotUtils:
    def filter_attribute_by_dict(self, target_attribute_dict=None):
        """
        Filters data by more than one attribute.
        """
        if bool(target_attribute_dict) != False:
            all_params = target_attribute_dict.keys()
            filtered_meta_pred_true = self.meta_pred_true.copy()
            for target_attribute in all_params:
                if self.is_list_numeric(self.meta_pred_true[target_attribute].values):
                    slider_min, slider_max = target_attribute_dict[target_attribute]
                    filtered_meta_pred_true = filtered_meta_pred_true[
                        self.meta_pred_true[target_attribute].between(
                            slider_min, slider_max
                        )
                    ]
                else:
                    target_value = target_attribute_dict[target_attribute]
                    filtered_meta_pred_true = filtered_meta_pred_true[
                        self.meta_pred_true[target_attribute] == target_value
                    ]
            return filtered_meta_pred_true
        else:
            return self.meta_pred_true

    def filter_attribute(self, target_attribute_dict):
        """
        Filters data by single attribute.
        """
        target_attribute = list(target_attribute_dict.keys())[0]
        target_value = target_attribute_dict[target_attribute]
        if self.is_list_numeric(self.meta_pred_true[target_attribute].values):
            slider_min, slider_max = target_attribute_dict[target_attribute]
            filtered_meta_pred_true = self.meta_pred_true[
                self.meta_pred_true[target_attribute].between(slider_min, slider_max)
            ]
        else:
            filtered_meta_pred_true = self.meta_pred_true[
                self.meta_pred_true[target_attribute] == target_value
            ]
        return filtered_meta_pred_true

    def multifilter_attribute(self, target_attributes_dict):
        """
        Filters data by more than one attribute.
        """
        all_params = target_attributes_dict.keys()
        filtered_meta_pred_true = self.meta_pred_true.copy()
        for target_attribute in all_params:
            if self.is_list_numeric(self.meta_pred_true[target_attribute].values):
                slider_min, slider_max = target_attributes_dict[target_attribute]
                filtered_meta_pred_true = filtered_meta_pred_true[
                    self.meta_pred_true[target_attribute].between(
                        slider_min, slider_max
                    )
                ]
            else:
                target_value = target_attributes_dict[target_attribute]
                filtered_meta_pred_true = filtered_meta_pred_true[
                    self.meta_pred_true[target_attribute] == target_value
                ]
        return filtered_meta_pred_true

    # Getters

    def get_classes(self):
        return self.class_order

    def get_predict_categorical(self, id_name):
        return self.meta_pred_true.loc[id_name]["pred_categorical"]

    def get_meta_df(self):
        return self.meta

    def get_entities(self):
        return self.meta.columns.tolist()

    def get_meta_with_results(self):
        return self.meta_pred_true.copy()

    # Typecheckers
    def is_list_numeric(self, x_list):
        return all(
            [
                isinstance(
                    i,
                    (
                        int,
                        float,
                        np.int,
                        np.int8,
                        np.int16,
                        np.int32,
                        np.float,
                        np.float16,
                        np.float32,
                        np.float64,
                    ),
                )
                for i in x_list
            ]
        )



import numpy as np
import pandas as pd
from sklearn.metrics import auc
import sklearn
from typing import Any, Dict, List, Optional, Union

from .confusion_matrix import ConfusionMatrix


class MostConfused:
    def __init__(self, class_mappings: Dict[int, str]) -> None:
        self.class_mappings = class_mappings
        self.class_order = [int(i) for i in list(class_mappings.keys())]

        self.confusion_matrix = ConfusionMatrix(class_mappings=class_mappings)

    def calculate_most_confused(
        self,
        true: Union[List[int], np.ndarray, pd.Series],
        pred_categorical: Union[List[int], np.ndarray, pd.Series],
    ) -> Dict[str, Any]:
        """Improve Explanation
        Calculates number of confusion for all classes on dataset.
        :return: confused_list_idxs , confused_list_values
        """
        confusion_matrix = self.confusion_matrix.calculate_confusion_matrix(
            true=true, pred_categorical=pred_categorical
        )
        confused_list_idxs = np.array([])
        confused_list_values = np.array([])
        for x in range(np.shape(confusion_matrix)[0]):
            for y in range(np.shape(confusion_matrix)[0]):
                if x != y:
                    confused_list_idxs = np.append(
                        confused_list_idxs, [self.class_order[x], self.class_order[y]]
                    )
                    confused_list_values = np.append(
                        confused_list_values, confusion_matrix[x, y]
                    )
        confused_list_idxs = confused_list_idxs.reshape(-1, 2)

        most_confused_df = pd.DataFrame(
            np.hstack((confused_list_idxs, np.reshape(confused_list_values, (-1, 1)))),
            columns=["true", "pred_categorical", "count"],
        )
        most_confused_df = most_confused_df.sort_values(by="count", ascending=False)
        most_confused_df.index = np.arange(len(most_confused_df.index))
        # Remove never confused samples
        most_confused_df = most_confused_df[most_confused_df["count"] > 0]
        return most_confused_df.to_dict()



import numpy as np
import pandas as pd

from ..metrics.top_losses import TopLosses


class PlotTopLosses:
    def __init__(self, pred, meta, meta_pred_true, loss, class_mappings):
        self.pred = pred
        self.loss = loss
        self.meta = meta
        self.meta_pred_true = meta_pred_true
        self.class_mappings = class_mappings

        self.top_losses_ = TopLosses(loss=self.loss, meta_pred_true=meta_pred_true)

    def top_losses(self, predicted_class=None, top_k=9):
        """
        Calculates minimum loss samples with appropriate information.
        :param self:
        :param predicted_class: Class of interest to calculate top losses.
        :param top_k: Number of most loss samples
        :return: payload_dict
        """

        # Check if overall top loss need to be observed.
        if predicted_class == "overall":
            predicted_class = None

        top_losses_class = self.top_losses_.calculate_top_losses(
            predicted_class, top_k
        ).head(top_k)
        top_loss_true_class = self.meta_pred_true.loc[top_losses_class.index]["true"]

        predicted_logits = self.pred.T.loc[top_losses_class.index]

        if predicted_class:
            logits_top_loss_predicted_class_probability = predicted_logits[
                predicted_class
            ]
        else:
            logits_top_loss_predicted_class_probability = predicted_logits.max(axis=1)

        top_loss_data = []
        i = 1  # FIX.
        for idx in top_losses_class.index:
            top_loss_data.append(
                {
                    "image_id": idx,
                    "rank": i,
                    "Ground Truth": self.class_mappings[
                        str(top_loss_true_class.loc[idx])
                    ],
                    "Prediction": self.class_mappings[
                        str(predicted_logits.loc[idx].idxmax())
                    ],
                    "Loss": float(top_losses_class.loc[idx]),
                    "Confidence": float(
                        logits_top_loss_predicted_class_probability.loc[idx]
                    ),
                    "Meta Data": self.meta.loc[idx].to_dict(),
                }
            )
            i += 1
        payload_dict = {"type": "image", "data": top_loss_data}
        return payload_dict
